import copy
from abc import abstractmethod
from typing import Callable, Dict, List, Optional, Tuple, get_args

import numpy as np
import pandas as pd
import torch
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.utils import get_infeasible_cost
from botorch.models.gpytorch import GPyTorchModel
from botorch.optim.initializers import gen_batch_initial_conditions
from botorch.optim.optimize import (
    optimize_acqf,
    optimize_acqf_discrete,
    optimize_acqf_list,
    optimize_acqf_mixed,
)
from torch import Tensor

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.enum import CategoricalEncodingEnum, CategoricalMethodEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    DiscreteInput,
    Input,
    TInputTransformSpecs,
)
from bofire.data_models.strategies.api import BotorchStrategy as DataModel
from bofire.data_models.strategies.api import (
    PolytopeSampler as PolytopeSamplerDataModel,
)
from bofire.data_models.surrogates.api import AnyTrainableSurrogate
from bofire.outlier_detection.outlier_detections import OutlierDetections
from bofire.strategies.predictives.predictive import PredictiveStrategy
from bofire.strategies.samplers.polytope import PolytopeSampler
from bofire.surrogates.botorch_surrogates import BotorchSurrogates
from bofire.utils.torch_tools import (
    get_initial_conditions_generator,
    get_linear_constraints,
    get_nchoosek_constraints,
    tkwargs,
)


def is_power_of_two(n):
    return (n != 0) and (n & (n - 1) == 0)


class BotorchStrategy(PredictiveStrategy):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.num_sobol_samples = data_model.num_sobol_samples
        self.num_restarts = data_model.num_restarts
        self.num_raw_samples = data_model.num_raw_samples
        self.descriptor_method = data_model.descriptor_method
        self.categorical_method = data_model.categorical_method
        self.discrete_method = data_model.discrete_method
        self.surrogate_specs = data_model.surrogate_specs
        if data_model.outlier_detection_specs is not None:
            self.outlier_detection_specs = OutlierDetections(
                data_model=data_model.outlier_detection_specs
            )
        else:
            self.outlier_detection_specs = None  # type: ignore
        self.min_experiments_before_outlier_check = (
            data_model.min_experiments_before_outlier_check
        )
        self.frequency_check = data_model.frequency_check
        self.frequency_hyperopt = data_model.frequency_hyperopt
        self.folds = data_model.folds
        self.surrogates = None
        torch.manual_seed(self.seed)

    model: Optional[GPyTorchModel] = None

    @property
    def input_preprocessing_specs(self) -> TInputTransformSpecs:
        return self.surrogate_specs.input_preprocessing_specs  # type: ignore

    @property
    def _features2idx(self) -> Dict[str, Tuple[int]]:
        features2idx, _ = self.domain.inputs._get_transform_info(
            self.input_preprocessing_specs
        )
        return features2idx

    @property
    def _features2names(self) -> Dict[str, Tuple[str]]:
        _, features2names = self.domain.inputs._get_transform_info(
            self.input_preprocessing_specs
        )
        return features2names

    def _fit(self, experiments: pd.DataFrame):
        """[summary]

        Args:
            transformed (pd.DataFrame): [description]
        """
        # perform outlier detection
        if self.outlier_detection_specs is not None:
            if (
                self.num_experiments >= self.min_experiments_before_outlier_check
                and self.num_experiments % self.frequency_check == 0
            ):
                experiments = self.outlier_detection_specs.detect(experiments)
        # perform hyperopt
        if (self.frequency_hyperopt > 0) and (
            self.num_experiments % self.frequency_hyperopt == 0
        ):
            # we have to import here to avoid circular imports
            from bofire.runners.hyperoptimize import hyperoptimize

            self.surrogate_specs.surrogates = [  # type: ignore
                hyperoptimize(
                    surrogate_data=surrogate_data,  # type: ignore
                    training_data=experiments,
                    folds=self.folds,
                )[0]
                if isinstance(surrogate_data, get_args(AnyTrainableSurrogate))
                else surrogate_data
                for surrogate_data in self.surrogate_specs.surrogates  # type: ignore
            ]

        # map the surrogate spec, we keep it here as attribute to be able to save/dump
        # the surrogate
        self.surrogates = BotorchSurrogates(data_model=self.surrogate_specs)  # type: ignore

        self.surrogates.fit(experiments)  # type: ignore
        self.model = self.surrogates.compatibilize(  # type: ignore
            inputs=self.domain.inputs,  # type: ignore
            outputs=self.domain.outputs,  # type: ignore
        )

    def _predict(self, transformed: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        # we are using self.model here for this purpose we have to take the transformed
        # input and further transform it to a torch tensor
        X = torch.from_numpy(transformed.values).to(**tkwargs)
        with torch.no_grad():
            posterior = self.model.posterior(X=X, observation_noise=True)  # type: ignore
        if len(posterior.mean.shape) == 2:
            preds = posterior.mean.cpu().detach().numpy()
            stds = np.sqrt(posterior.variance.cpu().detach().numpy())
        elif len(posterior.mean.shape) == 3:
            preds = posterior.mean.mean(dim=0).cpu().detach().numpy()
            stds = np.sqrt(posterior.variance.mean(dim=0).cpu().detach().numpy())
        else:
            raise ValueError("Wrong dimension of posterior mean. Expecting 2 or 3.")
        return preds, stds

    def calc_acquisition(
        self, candidates: pd.DataFrame, combined: bool = False
    ) -> np.ndarray:
        """Calculate the acqusition value for a set of experiments.

        Args:
            candidates (pd.DataFrame): Dataframe with experimentes for which the acqf value should be calculated.
            combined (bool, optional): If combined an acquisition value for the whole batch is calculated, else individual ones.
                Defaults to False.

        Returns:
            np.ndarray: Dataframe with the acquisition values.
        """
        acqf = self._get_acqfs(1)[0]

        transformed = self.domain.inputs.transform(  # type: ignore
            candidates, self.input_preprocessing_specs
        )
        X = torch.from_numpy(transformed.values).to(**tkwargs)
        if combined is False:
            X = X.unsqueeze(-2)

        with torch.no_grad():
            vals = acqf.forward(X).cpu().detach().numpy()

        return vals

    def _setup_ask(self):
        """Generates argument that can by passed to one of botorch's `optimize_acqf` method."""
        num_categorical_features = len(
            self.domain.get_features([CategoricalInput, DiscreteInput])
        )
        num_categorical_combinations = len(
            self.domain.inputs.get_categorical_combinations()
        )
        lower, upper = self.domain.inputs.get_bounds(
            specs=self.input_preprocessing_specs
        )
        bounds = torch.tensor([lower, upper]).to(**tkwargs)
        # setup nchooseks
        if len(self.domain.constraints.get(NChooseKConstraint)) == 0:
            ic_generator = None
            ic_gen_kwargs = {}
            nchooseks = None
        else:
            ic_generator = gen_batch_initial_conditions
            ic_gen_kwargs = {
                "generator": get_initial_conditions_generator(
                    strategy=PolytopeSampler(
                        data_model=PolytopeSamplerDataModel(domain=self.domain),
                    ),
                    transform_specs=self.input_preprocessing_specs,
                )
            }
            nchooseks = get_nchoosek_constraints(self.domain)
        # setup fixed features
        if (
            (num_categorical_features == 0)
            or (num_categorical_combinations == 1)
            or (
                all(
                    enc == CategoricalMethodEnum.FREE
                    for enc in [
                        self.categorical_method,
                        self.descriptor_method,
                        self.discrete_method,
                    ]
                )
            )
        ):
            fixed_features = self.get_fixed_features()
            fixed_features_list = None
        else:
            fixed_features = None
            fixed_features_list = self.get_categorical_combinations()
        return (
            bounds,
            ic_generator,
            ic_gen_kwargs,
            nchooseks,
            fixed_features,
            fixed_features_list,
        )

    def _postprocess_candidates(self, candidates: Tensor) -> pd.DataFrame:
        """Converts a tensor of candidates to a pandas Dataframe.

        Args:
            candidates (Tensor): Tensor of candidates returned from `optimize_acqf`.

        Returns:
            pd.DataFrame: Dataframe with candidates.
        """
        input_feature_keys = [
            item
            for key in self.domain.inputs.get_keys()
            for item in self._features2names[key]
        ]

        df_candidates = pd.DataFrame(
            data=candidates.detach().numpy(), columns=input_feature_keys
        )

        df_candidates = self.domain.inputs.inverse_transform(
            df_candidates, self.input_preprocessing_specs
        )

        preds = self.predict(df_candidates)
        return pd.concat((df_candidates, preds), axis=1)

    def _ask(self, candidate_count: int) -> pd.DataFrame:
        """[summary]

        Args:
            candidate_count (int, optional): [description]. Defaults to 1.

        Returns:
            pd.DataFrame: [description]
        """

        assert candidate_count > 0, "candidate_count has to be larger than zero."

        acqfs = self._get_acqfs(candidate_count)

        # we check here if we have a fully combinatorical search space
        if len(
            self.domain.inputs.get(includes=[DiscreteInput, CategoricalInput])
        ) == len(self.domain.inputs):
            if len(acqfs) > 1:
                raise NotImplementedError(
                    "Multiple Acqfs are currently not supported for purely combinatorical search spaces."
                )
            # generate the choices as pandas dataframe
            choices = pd.DataFrame.from_dict(
                [
                    {e[0]: e[1] for e in combi}  # type: ignore
                    for combi in self.domain.inputs.get_categorical_combinations()
                ]
            )
            # adding categorical features that are fixed
            for feat in self.domain.inputs.get_fixed():
                choices[feat.key] = feat.fixed_value()[0]  # type: ignore
            # compare the choices with the training data and remove all that are also part
            # of the training data
            merged = choices.merge(
                self.experiments[self.domain.inputs.get_keys()],
                on=list(choices.columns),
                how="left",
                indicator=True,
            )
            filtered_choices = merged[merged["_merge"] == "left_only"].copy()
            filtered_choices.drop(columns=["_merge"], inplace=True)

            # translate the filtered choice to torch
            t_choices = torch.from_numpy(
                self.domain.inputs.transform(
                    filtered_choices, specs=self.input_preprocessing_specs
                ).values
            ).to(**tkwargs)

            candidates, _ = optimize_acqf_discrete(
                acq_function=acqfs[0], q=candidate_count, unique=True, choices=t_choices
            )
            return self._postprocess_candidates(candidates=candidates)

        (
            bounds,
            ic_generator,
            ic_gen_kwargs,
            nchooseks,
            fixed_features,
            fixed_features_list,
        ) = self._setup_ask()

        if len(acqfs) > 1:
            candidates, _ = optimize_acqf_list(
                acq_function_list=acqfs,
                bounds=bounds,
                num_restarts=self.num_restarts,
                raw_samples=self.num_raw_samples,
                equality_constraints=get_linear_constraints(
                    domain=self.domain, constraint=LinearEqualityConstraint  # type: ignore
                ),
                inequality_constraints=get_linear_constraints(
                    domain=self.domain, constraint=LinearInequalityConstraint  # type: ignore
                ),
                nonlinear_inequality_constraints=nchooseks,
                fixed_features=fixed_features,
                fixed_features_list=fixed_features_list,
                ic_gen_kwargs=ic_gen_kwargs,
                ic_generator=ic_generator,
                options={"batch_limit": 5, "maxiter": 200},
            )
        else:
            if fixed_features_list:
                candidates, _ = optimize_acqf_mixed(
                    acq_function=acqfs[0],
                    bounds=bounds,
                    q=candidate_count,
                    num_restarts=self.num_restarts,
                    raw_samples=self.num_raw_samples,
                    equality_constraints=get_linear_constraints(
                        domain=self.domain, constraint=LinearEqualityConstraint  # type: ignore
                    ),
                    inequality_constraints=get_linear_constraints(
                        domain=self.domain, constraint=LinearInequalityConstraint  # type: ignore
                    ),
                    nonlinear_inequality_constraints=nchooseks,
                    fixed_features_list=fixed_features_list,
                    ic_generator=ic_generator,
                    ic_gen_kwargs=ic_gen_kwargs,
                )
            else:
                candidates, _ = optimize_acqf(
                    acq_function=acqfs[0],
                    bounds=bounds,
                    q=candidate_count,
                    num_restarts=self.num_restarts,
                    raw_samples=self.num_raw_samples,
                    equality_constraints=get_linear_constraints(
                        domain=self.domain, constraint=LinearEqualityConstraint  # type: ignore
                    ),
                    inequality_constraints=get_linear_constraints(
                        domain=self.domain, constraint=LinearInequalityConstraint  # type: ignore
                    ),
                    fixed_features=fixed_features,
                    nonlinear_inequality_constraints=nchooseks,
                    return_best_only=True,
                    ic_generator=ic_generator,  # type: ignore
                    **ic_gen_kwargs,  # type: ignore
                )
        return self._postprocess_candidates(candidates=candidates)

    def _tell(self) -> None:
        pass

    @abstractmethod
    def _get_acqfs(self, n: int) -> List[AcquisitionFunction]:
        pass

    def get_fixed_features(self):
        """provides the values of all fixed features

        Raises:
            NotImplementedError: [description]

        Returns:
            fixed_features (dict): Dictionary of fixed features, keys are the feature indices, values the transformed feature values
        """
        fixed_features = {}
        features2idx = self._features2idx

        for _, feat in enumerate(self.domain.get_features(Input)):
            if feat.fixed_value() is not None:  # type: ignore
                fixed_values = feat.fixed_value(transform_type=self.input_preprocessing_specs.get(feat.key))  # type: ignore
                for j, idx in enumerate(features2idx[feat.key]):
                    fixed_features[idx] = fixed_values[j]  # type: ignore

        # in case the optimization method is free and not allowed categories are present
        # one has to fix also them, this is abit of double work as it should be also reflected
        # in the bounds but helps to make it safer

        if (
            self.categorical_method == CategoricalMethodEnum.FREE
            and CategoricalEncodingEnum.ONE_HOT
            in list(self.input_preprocessing_specs.values())
        ):
            # for feat in self.get_true_categorical_features():
            for feat in [
                self.domain.inputs.get_by_key(featkey)
                for featkey in self.domain.inputs.get_keys(CategoricalInput)
                if self.input_preprocessing_specs[featkey]
                == CategoricalEncodingEnum.ONE_HOT
            ]:
                assert isinstance(feat, CategoricalInput)
                if feat.is_fixed() is False:
                    for cat in feat.get_forbidden_categories():
                        transformed = feat.to_onehot_encoding(pd.Series([cat]))
                        # we fix those indices to zero where one has a 1 as response from the transformer
                        for j, idx in enumerate(features2idx[feat.key]):
                            if transformed.values[0, j] == 1.0:
                                fixed_features[idx] = 0
        # for the descriptor ones
        if (
            self.descriptor_method == CategoricalMethodEnum.FREE
            and CategoricalEncodingEnum.DESCRIPTOR
            in list(self.input_preprocessing_specs.values())
        ):
            # for feat in self.get_true_categorical_features():
            for feat in [
                self.domain.inputs.get_by_key(featkey)
                for featkey in self.domain.inputs.get_keys(CategoricalDescriptorInput)
                if self.input_preprocessing_specs[featkey]
                == CategoricalEncodingEnum.DESCRIPTOR
            ]:
                assert isinstance(feat, CategoricalDescriptorInput)
                if feat.is_fixed() is False:
                    lower, upper = feat.get_bounds(CategoricalEncodingEnum.DESCRIPTOR)
                    for j, idx in enumerate(features2idx[feat.key]):
                        if lower[j] == upper[j]:
                            fixed_features[idx] = lower[j]
        return fixed_features

    def get_categorical_combinations(self):
        """provides all possible combinations of fixed values

        Returns:
            list_of_fixed_features List[dict]: Each dict contains a combination of fixed values
        """
        fixed_basis = self.get_fixed_features()

        methods = [
            self.descriptor_method,
            self.discrete_method,
            self.categorical_method,
        ]

        if all(m == CategoricalMethodEnum.FREE for m in methods):
            return [{}]
        else:
            include = []
            exclude = None

            if self.discrete_method == CategoricalMethodEnum.EXHAUSTIVE:
                include.append(DiscreteInput)

            if self.categorical_method == CategoricalMethodEnum.EXHAUSTIVE:
                include.append(CategoricalInput)
                exclude = CategoricalDescriptorInput

            if self.descriptor_method == CategoricalMethodEnum.EXHAUSTIVE:
                include.append(CategoricalDescriptorInput)
                exclude = None

        if not include:
            include = None

        combos = self.domain.inputs.get_categorical_combinations(
            include=(include if include else Input), exclude=exclude
        )
        # now build up the fixed feature list
        if len(combos) == 1:
            return [fixed_basis]
        else:
            features2idx = self._features2idx
            list_of_fixed_features = []

            for combo in combos:
                fixed_features = copy.deepcopy(fixed_basis)

                for pair in combo:
                    feat, val = pair
                    feature = self.domain.get_feature(feat)
                    if (
                        isinstance(feature, CategoricalDescriptorInput)
                        and self.input_preprocessing_specs[feat]
                        == CategoricalEncodingEnum.DESCRIPTOR
                    ):
                        index = feature.categories.index(val)

                        for j, idx in enumerate(features2idx[feat]):
                            fixed_features[idx] = feature.values[index][j]

                    elif isinstance(feature, CategoricalInput):
                        # it has to be onehot in this case
                        transformed = feature.to_onehot_encoding(pd.Series([val]))
                        for j, idx in enumerate(features2idx[feat]):
                            fixed_features[idx] = transformed.values[0, j]

                    elif isinstance(feature, DiscreteInput):
                        fixed_features[features2idx[feat][0]] = val

                list_of_fixed_features.append(fixed_features)
        return list_of_fixed_features

    def has_sufficient_experiments(
        self,
    ) -> bool:
        if self.experiments is None:
            return False
        if (
            len(
                self.domain.outputs.preprocess_experiments_all_valid_outputs(
                    experiments=self.experiments
                )
            )
            > 1
        ):
            return True
        return False

    def get_acqf_input_tensors(self):
        experiments = self.domain.outputs.preprocess_experiments_all_valid_outputs(
            self.experiments
        )

        # TODO: should this be selectable?
        clean_experiments = experiments.drop_duplicates(
            subset=[var.key for var in self.domain.get_features(Input)],
            keep="first",
            inplace=False,
        )

        transformed = self.domain.inputs.transform(
            clean_experiments, self.input_preprocessing_specs
        )
        X_train = torch.from_numpy(transformed.values).to(**tkwargs)

        if self.candidates is not None:
            transformed_candidates = self.domain.inputs.transform(
                self.candidates, self.input_preprocessing_specs
            )
            X_pending = torch.from_numpy(transformed_candidates.values).to(**tkwargs)
        else:
            X_pending = None

        return X_train, X_pending

    def get_infeasible_cost(
        self, objective: Callable[[Tensor, Tensor], Tensor], n_samples=128
    ) -> Tensor:
        X_train, X_pending = self.get_acqf_input_tensors()
        sampler = PolytopeSampler(
            data_model=PolytopeSamplerDataModel(domain=self.domain)
        )
        samples = sampler.ask(candidate_count=n_samples, return_all=False)
        # we need to transform the samples
        transformed_samples = torch.from_numpy(
            self.domain.inputs.transform(samples, self.input_preprocessing_specs).values
        ).to(**tkwargs)
        X = (
            torch.cat((X_train, X_pending, transformed_samples))
            if X_pending is not None
            else torch.cat((X_train, transformed_samples))
        )
        return get_infeasible_cost(
            X=X, model=self.model, objective=objective  # type: ignore
        )



import base64
import warnings
from typing import Callable, List, Tuple, Union

try:
    import cloudpickle
except ModuleNotFoundError:
    warnings.warn(
        "Cloudpickle is not available. CustomSoboStrategy's `f` cannot be dumped or loaded."
    )

import torch
from botorch.acquisition import get_acquisition_function
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.objective import ConstrainedMCObjective, GenericMCObjective
from botorch.models.gpytorch import GPyTorchModel

from bofire.data_models.acquisition_functions.api import qLogNEI, qNEI, qPI, qSR, qUCB
from bofire.data_models.objectives.api import ConstrainedObjective, Objective
from bofire.data_models.strategies.api import AdditiveSoboStrategy as AdditiveDataModel
from bofire.data_models.strategies.api import CustomSoboStrategy as CustomDataModel
from bofire.data_models.strategies.api import (
    MultiplicativeSoboStrategy as MultiplicativeDataModel,
)
from bofire.data_models.strategies.predictives.sobo import SoboBaseStrategy as DataModel
from bofire.strategies.predictives.botorch import BotorchStrategy
from bofire.utils.torch_tools import (
    get_additive_botorch_objective,
    get_custom_botorch_objective,
    get_multiplicative_botorch_objective,
    get_objective_callable,
    get_output_constraints,
    tkwargs,
)


class SoboStrategy(BotorchStrategy):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.acquisition_function = data_model.acquisition_function

    def _get_acqfs(self, n) -> List[AcquisitionFunction]:
        assert self.is_fitted is True, "Model not trained."

        X_train, X_pending = self.get_acqf_input_tensors()

        (
            objective_callable,
            constraint_callables,
            etas,
        ) = self._get_objective_and_constraints()

        assert self.model is not None

        acqf = get_acquisition_function(
            self.acquisition_function.__class__.__name__,
            self.model,
            objective_callable,
            X_observed=X_train,
            X_pending=X_pending,
            constraints=constraint_callables,
            mc_samples=self.num_sobol_samples,
            beta=self.acquisition_function.beta
            if isinstance(self.acquisition_function, qUCB)
            else 0.2,
            tau=self.acquisition_function.tau
            if isinstance(self.acquisition_function, qPI)
            else 1e-3,
            eta=torch.tensor(etas).to(**tkwargs),
            cache_root=True if isinstance(self.model, GPyTorchModel) else False,
            prune_baseline=self.acquisition_function.prune_baseline
            if isinstance(self.acquisition_function, (qNEI, qLogNEI))
            else True,
        )
        return [acqf]

    def _get_objective_and_constraints(
        self,
    ) -> Tuple[
        Union[GenericMCObjective, ConstrainedMCObjective],
        Union[List[Callable[[torch.Tensor], torch.Tensor]], None],
        Union[List, float],
    ]:
        try:
            target_feature = self.domain.outputs.get_by_objective(
                excludes=ConstrainedObjective
            )[0]
        except IndexError:
            target_feature = self.domain.outputs.get_by_objective(includes=Objective)[0]
        target_index = self.domain.outputs.get_keys().index(target_feature.key)
        objective_callable = get_objective_callable(
            idx=target_index, objective=target_feature.objective
        )

        # get the constraints
        if (len(self.domain.outputs.get_by_objective(ConstrainedObjective)) > 0) and (
            len(self.domain.outputs.get_by_objective(Objective)) > 1
        ):
            constraint_callables, etas = get_output_constraints(
                outputs=self.domain.outputs
            )
        else:
            constraint_callables, etas = None, 1e-3

        # special cases of qUCB and qSR do not work with separate constraints
        if (isinstance(self.acquisition_function, (qSR, qUCB))) and (
            constraint_callables is not None
        ):
            return (
                ConstrainedMCObjective(
                    objective=objective_callable,  # type: ignore
                    constraints=constraint_callables,
                    eta=torch.tensor(etas).to(**tkwargs),
                    infeasible_cost=self.get_infeasible_cost(
                        objective=objective_callable
                    ),
                ),
                None,
                1e-3,
            )

        # return regular objective
        return (
            GenericMCObjective(objective=objective_callable),
            constraint_callables,
            etas,
        )


class AdditiveSoboStrategy(SoboStrategy):
    def __init__(
        self,
        data_model: AdditiveDataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.use_output_constraints = data_model.use_output_constraints

    def _get_objective_and_constraints(
        self,
    ) -> Tuple[
        Union[GenericMCObjective, ConstrainedMCObjective],
        Union[List[Callable[[torch.Tensor], torch.Tensor]], None],
        Union[List, float],
    ]:
        # get the constraints
        if (
            (len(self.domain.outputs.get_by_objective(ConstrainedObjective)) > 0)
            and (len(self.domain.outputs.get_by_objective(Objective)) > 1)
            and self.use_output_constraints
        ):
            constraint_callables, etas = get_output_constraints(
                outputs=self.domain.outputs
            )
        else:
            constraint_callables, etas = None, 1e-3
        # TODO: test this
        if self.use_output_constraints:
            objective_callable = get_additive_botorch_objective(
                outputs=self.domain.outputs, exclude_constraints=True
            )

            # special cases of qUCB and qSR do not work with separate constraints
            if isinstance(self.acquisition_function, (qSR, qUCB)):
                return (
                    ConstrainedMCObjective(
                        objective=objective_callable,  # type: ignore
                        constraints=constraint_callables,
                        eta=torch.tensor(etas).to(**tkwargs),
                        infeasible_cost=self.get_infeasible_cost(
                            objective=objective_callable
                        ),
                    ),
                    None,
                    1e-3,
                )
            else:
                return (
                    GenericMCObjective(objective=objective_callable),
                    constraint_callables,
                    etas,
                )

        # we absorb all constraints into the objective
        return (
            GenericMCObjective(
                objective=get_additive_botorch_objective(  # type: ignore
                    outputs=self.domain.outputs, exclude_constraints=False
                )
            ),
            constraint_callables,
            etas,
        )


class MultiplicativeSoboStrategy(SoboStrategy):
    def __init__(
        self,
        data_model: MultiplicativeDataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)

    def _get_objective_and_constraints(
        self,
    ) -> Tuple[
        GenericMCObjective,
        Union[List[Callable[[torch.Tensor], torch.Tensor]], None],
        Union[List, float],
    ]:
        # we absorb all constraints into the objective
        return (
            GenericMCObjective(
                objective=get_multiplicative_botorch_objective(  # type: ignore
                    outputs=self.domain.outputs
                )
            ),
            None,
            1e-3,
        )


class CustomSoboStrategy(SoboStrategy):
    def __init__(
        self,
        data_model: CustomDataModel,
        **kwargs,
    ):
        super().__init__(data_model=data_model, **kwargs)
        self.use_output_constraints = data_model.use_output_constraints
        if data_model.dump is not None:
            self.loads(data_model.dump)
        else:
            self.f = None

    def _get_objective_and_constraints(
        self,
    ) -> Tuple[
        Union[GenericMCObjective, ConstrainedMCObjective],
        Union[List[Callable[[torch.Tensor], torch.Tensor]], None],
        Union[List, float],
    ]:
        if self.f is None:
            raise ValueError("No function has been provided for the strategy")
        # get the constraints
        if (
            (len(self.domain.outputs.get_by_objective(ConstrainedObjective)) > 0)
            and (len(self.domain.outputs.get_by_objective(Objective)) > 1)
            and self.use_output_constraints
        ):
            constraint_callables, etas = get_output_constraints(
                outputs=self.domain.outputs
            )
        else:
            constraint_callables, etas = None, 1e-3

        if self.use_output_constraints:
            objective_callable = get_custom_botorch_objective(
                outputs=self.domain.outputs, f=self.f, exclude_constraints=True
            )
            # special cases of qUCB and qSR do not work with separate constraints
            if isinstance(self.acquisition_function, (qSR, qUCB)):
                return (
                    ConstrainedMCObjective(
                        objective=objective_callable,  # type: ignore
                        constraints=constraint_callables,
                        eta=torch.tensor(etas).to(**tkwargs),
                        infeasible_cost=self.get_infeasible_cost(
                            objective=objective_callable
                        ),
                    ),
                    None,
                    1e-3,
                )
            else:
                return (
                    GenericMCObjective(objective=objective_callable),
                    constraint_callables,
                    etas,
                )

        # we absorb all constraints into the objective
        return (
            GenericMCObjective(
                objective=get_custom_botorch_objective(
                    outputs=self.domain.outputs, f=self.f, exclude_constraints=False
                )
            ),
            constraint_callables,
            etas,
        )

    def dumps(self) -> str:
        """Dumps the function to a string via pickle as this is not directly json serializable."""
        if self.f is None:
            raise ValueError("No function has been provided for the strategy")
        f_bytes_dump = cloudpickle.dumps(self.f)
        return base64.b64encode(f_bytes_dump).decode()

    def loads(self, data: str):
        """Loads the function from a base64 encoded pickle bytes object and writes it to the `model` attribute."""
        f_bytes_load = base64.b64decode(data.encode())
        self.f = cloudpickle.loads(f_bytes_load)



from abc import ABC
from typing import List

import pandas as pd

from bofire.data_models.outlier_detection.api import OutlierDetections as DataModel
from bofire.outlier_detection.api import (  # noqa: F401
    IterativeTrimming,
    OutlierDetection,
)
from bofire.outlier_detection.api import map as map_outlier

AnyOutlierDetector = IterativeTrimming


class OutlierDetections(ABC):
    detectors: List[AnyOutlierDetector]

    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        self.detectors = [map_outlier(model) for model in data_model.detectors]  # type: ignore

    def detect(self, experiments: pd.DataFrame) -> pd.DataFrame:
        for outlier_model in self.detectors:  # type: ignore
            experiments = outlier_model.detect(experiments)

        return experiments



import random
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Type


class Spec:
    """Spec <spec> for class <cls>."""

    def __init__(self, cls: Type, spec: Callable[[], dict]):
        self.cls = cls
        self.spec = spec

    def obj(self, **kwargs) -> Any:
        """Create and return an instance of <cls>."""

        return self.cls(**{**self.spec(), **kwargs})

    def typed_spec(self) -> dict:
        """Return the <spec>, extended by {'type': <cls>.__name__}."""
        return {
            **self.spec(),
            "type": self.cls.__name__,
        }

    def __str__(self):
        return f"{self.cls.__name__}: {self.spec()}"

    def __repr__(self):
        return str(self)


class Invalidator(ABC):
    """Invalidator rule to invalidate a given spec."""

    @abstractmethod
    def invalidate(self, spec: Spec) -> List[Spec]:
        """Return a list of invalidated specs.

        If this invalidator is not applicable to the specified
        spec, an empty list is returned."""

        pass


class Overwrite(Invalidator):
    """Overwrite properties if the key is contained in the spec."""

    def __init__(self, key: str, overwrites: List[dict]):
        self.key = key
        self.overwrites = overwrites

    def invalidate(self, spec: Spec) -> List[Spec]:
        data: dict = spec.spec()
        if self.key not in data:
            return []
        return [
            Spec(spec.cls, lambda data=data, overwrite=overwrite: {**data, **overwrite})
            for overwrite in self.overwrites
        ]


class Specs:
    """Collection of valid and invalid specs.

    In the init, only <invalidators> must be provided.
    Valid specs are added via the <add_valid> method.
    Invalid specs can auomatically be added as part of this method."""

    def __init__(self, invalidators: List[Invalidator]):
        self.invalidators = invalidators
        self.valids: List[Spec] = []
        self.invalids: List[Spec] = []

    def _get_spec(self, specs: List[Spec], cls: Type = None, exact: bool = True):
        if cls is not None:
            if exact:
                specs = [s for s in specs if s.cls == cls]
            else:
                specs = [s for s in specs if issubclass(s.cls, cls)]
        if len(specs) == 0 and cls is None:
            raise TypeError("no spec found")
        elif len(specs) == 0:
            raise TypeError(f"no spec of type {cls.__name__} found")
        return random.choice(specs)

    def valid(self, cls: Type = None, exact: bool = True) -> Spec:
        """Return a valid spec.

        If <cls> is provided, the list of all valid specs is filtered by it.
        If no spec (with the specified class) exists, a TypeError is raised.
        If more than one spec exist, a random one is returned."""

        return self._get_spec(self.valids, cls, exact)

    def invalid(self, cls: Type = None, exact: bool = True) -> Spec:
        """Return an invalid spec.

        If <cls> is provided, the list of all invalid specs is filtered by it.
        If no spec (with the specified class) exists, a TypeError is raised.
        If more than one spec exist, a random one is returned."""

        return self._get_spec(self.invalids, cls, exact)

    def add_valid(
        self, cls: Type, spec: Callable[[], dict], add_invalids: bool = True
    ) -> Spec:
        """Add a new valid spec to the list.

        If <add_invalids> is True (default), invalid specs are generated using the
        rules provided in <invalidators>."""

        spec_ = Spec(cls, spec)
        self.valids.append(spec_)
        if add_invalids:
            for invalidator in self.invalidators:
                self.invalids += invalidator.invalidate(spec_)
        return spec_

    def add_invalid(self, cls: Type, spec: Callable[[], dict]) -> Spec:
        """Add a new invalid spec to the list."""

        spec_ = Spec(cls, spec)
        self.invalids.append(spec_)
        return spec_



import random

import bofire.data_models.surrogates.api as models
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
    MolecularInput,
)
from bofire.data_models.kernels.api import (
    HammondDistanceKernel,
    MaternKernel,
    ScaleKernel,
    TanimotoKernel,
)
from bofire.data_models.molfeatures.api import Fingerprints
from bofire.data_models.priors.api import (
    BOTORCH_LENGTHCALE_PRIOR,
    BOTORCH_NOISE_PRIOR,
    BOTORCH_SCALE_PRIOR,
)
from bofire.data_models.surrogates.api import (
    MeanAggregation,
    ScalerEnum,
    SumAggregation,
)
from bofire.data_models.surrogates.single_task_gp import SingleTaskGPHyperconfig
from tests.bofire.data_models.specs.features import specs as features
from tests.bofire.data_models.specs.specs import Specs

specs = Specs([])

specs.add_valid(
    models.SingleTaskGPSurrogate,
    lambda: {
        "inputs": Inputs(
            features=[
                ContinuousInput(key="a", bounds=(0, 1)),
                ContinuousInput(key="b", bounds=(0, 1)),
            ]
        ),
        "outputs": Outputs(
            features=[
                features.valid(ContinuousOutput).obj(),
            ]
        ),
        "kernel": ScaleKernel(
            base_kernel=MaternKernel(
                ard=True, nu=2.5, lengthscale_prior=BOTORCH_LENGTHCALE_PRIOR()
            ),
            outputscale_prior=BOTORCH_SCALE_PRIOR(),
        ),
        "aggregations": [
            random.choice(
                [
                    SumAggregation(features=["a", "b"]),
                    MeanAggregation(features=["a", "b"]),
                ]
            )
        ],
        "scaler": ScalerEnum.NORMALIZE,
        "noise_prior": BOTORCH_NOISE_PRIOR(),
        "input_preprocessing_specs": {},
        "dump": None,
        "hyperconfig": SingleTaskGPHyperconfig(),
    },
)

specs.add_valid(
    models.MixedSingleTaskGPSurrogate,
    lambda: {
        "inputs": Inputs(
            features=[
                features.valid(ContinuousInput).obj(),
            ]
            + [CategoricalInput(key="cat1", categories=["a", "b", "c"])]
        ),
        "outputs": Outputs(
            features=[
                features.valid(ContinuousOutput).obj(),
            ]
        ),
        "aggregations": None,
        "continuous_kernel": MaternKernel(ard=True, nu=random.random()),
        "categorical_kernel": HammondDistanceKernel(ard=True),
        "scaler": ScalerEnum.NORMALIZE,
        "input_preprocessing_specs": {"cat1": CategoricalEncodingEnum.ONE_HOT},
        "dump": None,
        "hyperconfig": None,
    },
)
specs.add_valid(
    models.SingleTaskGPSurrogate,
    lambda: {
        "inputs": Inputs(
            features=[
                features.valid(ContinuousInput).obj(),
            ]
        ),
        "outputs": Outputs(
            features=[
                features.valid(ContinuousOutput).obj(),
            ]
        ),
        "kernel": ScaleKernel(
            base_kernel=MaternKernel(
                ard=True, nu=2.5, lengthscale_prior=BOTORCH_LENGTHCALE_PRIOR()
            ),
            outputscale_prior=BOTORCH_SCALE_PRIOR(),
        ),
        "aggregations": None,
        "scaler": ScalerEnum.NORMALIZE,
        "noise_prior": BOTORCH_NOISE_PRIOR(),
        "input_preprocessing_specs": {},
        "dump": None,
        "hyperconfig": SingleTaskGPHyperconfig(),
    },
)
specs.add_valid(
    models.RandomForestSurrogate,
    lambda: {
        "inputs": Inputs(
            features=[
                features.valid(ContinuousInput).obj(),
            ]
        ),
        "outputs": Outputs(
            features=[
                features.valid(ContinuousOutput).obj(),
            ]
        ),
        "aggregations": None,
        "input_preprocessing_specs": {},
        "n_estimators": 100,
        "criterion": "squared_error",
        "max_depth": None,
        "min_samples_split": 2,
        "min_samples_leaf": 1,
        "min_weight_fraction_leaf": 0.0,
        "max_features": 1,
        "max_leaf_nodes": None,
        "min_impurity_decrease": 0.0,
        "bootstrap": True,
        "oob_score": False,
        "random_state": None,
        "ccp_alpha": 0.0,
        "max_samples": None,
        "dump": None,
        "hyperconfig": None,
    },
)
specs.add_valid(
    models.MLPEnsemble,
    lambda: {
        "inputs": Inputs(
            features=[
                features.valid(ContinuousInput).obj(),
            ]
        ),
        "outputs": Outputs(
            features=[
                features.valid(ContinuousOutput).obj(),
            ]
        ),
        "aggregations": None,
        "n_estimators": 2,
        "hidden_layer_sizes": (100,),
        "activation": "relu",
        "dropout": 0.0,
        "batch_size": 10,
        "n_epochs": 200,
        "lr": 1e-4,
        "weight_decay": 0.0,
        "subsample_fraction": 1.0,
        "shuffle": True,
        "scaler": ScalerEnum.NORMALIZE,
        "output_scaler": ScalerEnum.STANDARDIZE,
        "input_preprocessing_specs": {},
        "dump": None,
        "hyperconfig": None,
    },
)

specs.add_valid(
    models.XGBoostSurrogate,
    lambda: {
        "inputs": Inputs(
            features=[
                features.valid(ContinuousInput).obj(),
            ]
        ),
        "outputs": Outputs(
            features=[
                features.valid(ContinuousOutput).obj(),
            ]
        ),
        "aggregations": None,
        "n_estimators": 10,
        "max_depth": 6,
        "max_leaves": 0,
        "max_bin": 256,
        "grow_policy": "depthwise",
        "learning_rate": 0.3,
        "objective": "reg:squarederror",
        "booster": "gbtree",
        "n_jobs": 1,
        "gamma": 0.0,
        "min_child_weight": 1.0,
        "max_delta_step": 0.0,
        "subsample": 1.0,
        "sampling_method": "uniform",
        "colsample_bytree": 1.0,
        "colsample_bylevel": 1.0,
        "colsample_bynode": 1.0,
        "reg_alpha": 0.0,
        "reg_lambda": 1.0,
        "scale_pos_weight": 1,
        "random_state": None,
        "num_parallel_tree": 1,
        "input_preprocessing_specs": {},
        "dump": None,
        "hyperconfig": None,
    },
)
specs.add_valid(
    models.TanimotoGPSurrogate,
    lambda: {
        "inputs": Inputs(
            features=[
                MolecularInput(key="mol1"),
            ]
        ),
        "outputs": Outputs(
            features=[
                features.valid(ContinuousOutput).obj(),
            ]
        ),
        "kernel": ScaleKernel(
            base_kernel=TanimotoKernel(
                ard=True,
            ),
            outputscale_prior=BOTORCH_SCALE_PRIOR(),
        ),
        "aggregations": None,
        "scaler": ScalerEnum.NORMALIZE,
        "noise_prior": BOTORCH_NOISE_PRIOR(),
        "input_preprocessing_specs": {"mol1": Fingerprints(n_bits=32, bond_radius=3)},
        "dump": None,
        "hyperconfig": None,
    },
)



from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd

from bofire.data_models.domain.domain import is_numeric
from bofire.data_models.surrogates.api import Surrogate as DataModel
from bofire.surrogates.values import PredictedValue


class Surrogate(ABC):
    def __init__(
        self,
        data_model: DataModel,
    ):
        self.inputs = data_model.inputs
        self.outputs = data_model.outputs
        self.input_preprocessing_specs = data_model.input_preprocessing_specs
        if data_model.dump is not None:
            self.loads(data_model.dump)
        else:
            self.model = None

    @classmethod
    def from_spec(cls, data_model: DataModel) -> "Surrogate":
        return cls(data_model=data_model)

    @property
    def is_fitted(self) -> bool:
        """Return True if model is fitted, else False."""
        return self.model is not None

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        # check if model is fitted
        if not self.is_fitted:
            raise ValueError("Model is not fitted/available yet.")
        # validate
        X = self.inputs.validate_experiments(X, strict=False)
        # transform
        Xt = self.inputs.transform(X, self.input_preprocessing_specs)
        # make sure that all is float64
        for c in Xt.columns:
            Xt[c] = pd.to_numeric(Xt[c], errors="raise")
        # predict
        preds, stds = self._predict(Xt)
        # postprocess
        predictions = pd.DataFrame(
            data=np.hstack((preds, stds)),
            columns=["%s_pred" % featkey for featkey in self.outputs.get_keys()]
            + ["%s_sd" % featkey for featkey in self.outputs.get_keys()],
        )
        # validate
        self.validate_predictions(predictions=predictions)
        # return
        return predictions

    def validate_predictions(self, predictions: pd.DataFrame) -> pd.DataFrame:
        expected_cols = [
            f"{key}_{t}" for key in self.outputs.get_keys() for t in ["pred", "sd"]
        ]
        if sorted(predictions.columns) != sorted(expected_cols):
            raise ValueError(
                f"Predictions are ill-formatted. Expected: {expected_cols}, got: {list(predictions.columns)}."
            )
        # check that values are numeric
        if not is_numeric(predictions):
            raise ValueError("Not all values in predictions are numeric.")
        return predictions

    def to_predictions(
        self, predictions: pd.DataFrame
    ) -> Dict[str, List[PredictedValue]]:
        outputs = {key: [] for key in self.outputs.get_keys()}
        for _, row in predictions.iterrows():
            for key in self.outputs.get_keys():
                outputs[key].append(
                    PredictedValue(
                        predictedValue=row[f"{key}_pred"],
                        standardDeviation=row[f"{key}_sd"],
                    )
                )
        return outputs

    @abstractmethod
    def _predict(self, transformed_X: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        pass

    def dumps(self) -> str:
        """Dumps the actual model to a string as this is not directly json serializable."""
        if not self.is_fitted:
            raise ValueError("Model has to be fitted before dumping")
        self._prepare_for_dump()
        return self._dumps()

    @abstractmethod
    def _dumps(self) -> str:
        pass

    def _prepare_for_dump(self):
        """Prepares the model before the dump."""
        pass

    @abstractmethod
    def loads(self, data: str):
        """Loads the actual model from a string and writes it to the `model` attribute."""
        pass



import unittest

import numpy as np
import pandas as pd
import pytest

from bofire.data_models.constraints.api import NChooseKConstraint
from bofire.data_models.domain.api import Constraints, Domain, Inputs
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)

# NChooseKConstraint constraints 1
cc1a = NChooseKConstraint(
    features=["0", "1", "2", "3"], min_count=2, max_count=3, none_also_valid=True
)
cc2a = NChooseKConstraint(
    features=["2", "3", "4", "5"], min_count=1, max_count=2, none_also_valid=True
)

# NChooseKConstraint constraints 2
cc1b = NChooseKConstraint(
    features=["0", "1", "2", "3"], min_count=2, max_count=3, none_also_valid=False
)
cc2b = NChooseKConstraint(
    features=["2", "3", "4", "5"], min_count=1, max_count=2, none_also_valid=True
)

# NChooseKConstraint constraint 3
cc3 = [
    NChooseKConstraint(
        features=["0", "1", "2", "3"], min_count=2, max_count=3, none_also_valid=True
    )
]

# input features
continuous_inputs = []
for i in range(6):
    f = ContinuousInput(key=str(i), bounds=(0, 1))
    continuous_inputs.append(f)
categorical_feature = CategoricalInput(
    key="categorical_feature", categories=["c1", "c2"]
)
categorical_descriptor_feature = CategoricalDescriptorInput(
    key="categorical_descriptor_feature",
    categories=["cd1", "cd2"],
    descriptors=["d1", "d2"],
    values=[[1.0, 1.0], [2.0, 2.0]],
)

# output feature
outputs = [ContinuousOutput(key="y")]


"""
TEST CASES:

CASE 1: 6 continuous features, 2 overlapping NChooseKConstraint constraints, none_also_valid: True, True
CASE 2: 6 continuous features, 2 overlapping NChooseKConstraint constraints, none_also_valid: False, True
"""

# CASE 1
test_features_used_1 = [
    ["0", "1", "2"],
    ["0", "1", "3"],
    ["0", "1", "4"],
    ["0", "1", "5"],
    ["0", "1", "2", "4"],
    ["0", "1", "2", "5"],
    ["0", "1", "3", "4"],
    ["0", "1", "3", "5"],
    ["0", "1", "4", "5"],
    ["0", "1"],
    ["0", "2"],
    ["0", "2", "3"],
    ["0", "2", "4"],
    ["0", "2", "5"],
    ["0", "3"],
    ["0", "3", "4"],
    ["0", "3", "5"],
    ["1", "2"],
    ["1", "2", "3"],
    ["1", "2", "4"],
    ["1", "2", "5"],
    ["1", "3"],
    ["1", "3", "4"],
    ["1", "3", "5"],
    ["2", "3"],
    ["4"],
    ["5"],
    ["4", "5"],
    [],
]
test_features_unused_1 = []
for used in test_features_used_1:
    unused = [f for f in ["0", "1", "2", "3", "4", "5"] if f not in used]
    test_features_unused_1.append(unused)

# CASE 2
test_features_used_2 = [
    ["0", "1", "2"],
    ["0", "1", "3"],
    ["0", "1", "4"],
    ["0", "1", "5"],
    ["0", "1", "2", "4"],
    ["0", "1", "2", "5"],
    ["0", "1", "3", "4"],
    ["0", "1", "3", "5"],
    ["0", "1", "4", "5"],
    ["0", "1"],
    ["0", "2"],
    ["0", "2", "3"],
    ["0", "2", "4"],
    ["0", "2", "5"],
    ["0", "3"],
    ["0", "3", "4"],
    ["0", "3", "5"],
    ["1", "2"],
    ["1", "2", "3"],
    ["1", "2", "4"],
    ["1", "2", "5"],
    ["1", "3"],
    ["1", "3", "4"],
    ["1", "3", "5"],
    ["2", "3"],
]
test_features_unused_2 = []
for used in test_features_used_2:
    unused = [f for f in ["0", "1", "2", "3", "4", "5"] if f not in used]
    test_features_unused_2.append(unused)

# experiments
experiments = pd.DataFrame(
    np.random.uniform(size=(24, 7)), columns=["0", "1", "2", "3", "4", "5", "y"]
)
experiments["categorical_feature"] = ["c1"] * 12 + ["c2"] * 12
experiments["categorical_descriptor_feature"] = (["cd1"] * 6 + ["cd2"] * 6) * 2
experiments["valid_y"] = 1


##### LIST OF TASTE CASES #####

test_cases = []

# CASE 1
test_case = {}
domain = Domain.from_lists(
    inputs=continuous_inputs,
    outputs=outputs,
    constraints=[cc1a, cc2a],
)
test_case["domain"] = domain
test_case["experiments"] = experiments
test_case["descriptor_method"] = None
test_case["categorical_method"] = None
test_case["descriptor_encoding"] = None
test_case["categorical_encoding"] = None
test_case["test_features_used"] = test_features_used_1
test_case["test_features_unused"] = test_features_unused_1
test_cases.append(test_case)

# CASE 2
test_case = {}
domain = Domain.from_lists(
    continuous_inputs,
    outputs,
    constraints=[cc1b, cc2b],
)
test_case["domain"] = domain
test_case["experiments"] = experiments
test_case["descriptor_method"] = None
test_case["categorical_method"] = None
test_case["descriptor_encoding"] = None
test_case["categorical_encoding"] = None
test_case["test_features_used"] = test_features_used_2
test_case["test_features_unused"] = test_features_unused_2
test_cases.append(test_case)


@pytest.mark.parametrize("test_case", test_cases)
def test_nchoosek_combinations_completeness(test_case):
    domain = test_case["domain"]
    features_used, features_unused = domain.get_nchoosek_combinations(exhaustive=True)
    for features in test_case["test_features_used"]:
        assert features in features_used
    for features in test_case["test_features_unused"]:
        assert features in features_unused


def test_nchoosek_combinations_nonexhaustive():
    domain = Domain(
        inputs=Inputs(
            features=[ContinuousInput(key=f"if{i+1}", bounds=(0, 1)) for i in range(6)]
        ),
        constraints=Constraints(
            constraints=[
                NChooseKConstraint(
                    features=[f"if{i+1}" for i in range(4)],
                    min_count=0,
                    max_count=2,
                    none_also_valid=True,
                )
            ]
        ),
    )
    used, unused = domain.get_nchoosek_combinations(exhaustive=False)
    expected_used = [
        ["if1", "if2"],
        ["if1", "if3"],
        ["if1", "if4"],
        ["if2", "if3"],
        ["if2", "if4"],
        ["if3", "if4"],
    ]
    expected_unused = [
        ["if3", "if4"],
        ["if2", "if4"],
        ["if2", "if3"],
        ["if1", "if4"],
        ["if1", "if3"],
        ["if1", "if2"],
    ]
    # print(combos, expected_combos)
    c = unittest.TestCase()
    c.assertCountEqual(used, expected_used)
    c.assertCountEqual(unused, expected_unused)



from typing import Annotated, Literal, Sequence

from pydantic import Field, validator

from bofire.data_models.surrogates.botorch import BotorchSurrogate
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.data_models.surrogates.trainable import TrainableSurrogate


class MLPEnsemble(BotorchSurrogate, TrainableSurrogate):
    type: Literal["MLPEnsemble"] = "MLPEnsemble"
    n_estimators: Annotated[int, Field(ge=1)] = 5
    hidden_layer_sizes: Sequence = (100,)
    activation: Literal["relu", "logistic", "tanh"] = "relu"
    dropout: Annotated[float, Field(ge=0.0)] = 0.0
    batch_size: Annotated[int, Field(ge=1)] = 10
    n_epochs: Annotated[int, Field(ge=1)] = 200
    lr: Annotated[float, Field(gt=0.0)] = 1e-4
    weight_decay: Annotated[float, Field(ge=0.0)] = 0.0
    subsample_fraction: Annotated[float, Field(gt=0.0)] = 1.0
    shuffle: bool = True
    scaler: ScalerEnum = ScalerEnum.NORMALIZE
    output_scaler: ScalerEnum = ScalerEnum.STANDARDIZE

    @validator("output_scaler")
    def validate_output_scaler(cls, output_scaler):
        """validates that output_scaler is a valid type

        Args:
            output_scaler (ScalerEnum): Scaler used to transform the output

        Raises:
            ValueError: when ScalerEnum.NORMALIZE is used

        Returns:
            ScalerEnum: Scaler used to transform the output
        """
        if output_scaler == ScalerEnum.NORMALIZE:
            raise ValueError("Normalize is not supported as an output transform.")

        return output_scaler



from typing import Union

from bofire.data_models.objectives.identity import (
    IdentityObjective,
    MaximizeObjective,
    MinimizeObjective,
)
from bofire.data_models.objectives.objective import Objective
from bofire.data_models.objectives.sigmoid import (
    MaximizeSigmoidObjective,
    MinimizeSigmoidObjective,
    SigmoidObjective,
)
from bofire.data_models.objectives.target import (
    CloseToTargetObjective,
    ConstrainedObjective,
    TargetObjective,
)

AbstractObjective = Union[
    Objective,
    IdentityObjective,
    SigmoidObjective,
    ConstrainedObjective,
]

AnyConstraintObjective = Union[
    MaximizeSigmoidObjective,
    MinimizeSigmoidObjective,
    TargetObjective,
]

AnyRealObjective = Union[MaximizeObjective, MinimizeObjective, CloseToTargetObjective]

AnyObjective = Union[
    MaximizeObjective,
    MinimizeObjective,
    MaximizeSigmoidObjective,
    MinimizeSigmoidObjective,
    TargetObjective,
    CloseToTargetObjective,
]



from abc import abstractmethod
from copy import deepcopy
from typing import Type

import numpy as np
import pandas as pd
import torch
from formulaic import Formula
from torch import Tensor

from bofire.data_models.domain.api import Domain
from bofire.strategies.enum import OptimalityCriterionEnum
from bofire.utils.torch_tools import tkwargs


class Objective:
    def __init__(
        self,
        domain: Domain,
        model: Formula,
        n_experiments: int,
        delta: float = 1e-6,
    ) -> None:
        """
        Args:
            domain (Domain): A domain defining the DoE domain together with model_type.
            model_type (str or Formula): A formula containing all model terms.
            n_experiments (int): Number of experiments
            delta (float): A regularization parameter for the information matrix. Default value is 1e-3.

        """

        self.model = deepcopy(model)
        self.domain = deepcopy(domain)
        self.n_experiments = n_experiments
        self.delta = delta

        self.vars = self.domain.inputs.get_keys()
        self.n_vars = len(self.domain.inputs)

        self.model_terms = list(np.array(model, dtype=str))
        self.n_model_terms = len(self.model_terms)

        # terms for model jacobian
        self.terms_jacobian_t = []
        for var in self.vars:
            _terms = [
                str(term).replace(":", "*") + f" + 0 * {self.vars[0]}"
                for term in model.differentiate(var, use_sympy=True)
            ]  # 0*vars[0] added to make sure terms are evaluated as series, not as number
            terms = "["
            for t in _terms:
                terms += t + ", "
            terms = terms[:-1] + "]"

            self.terms_jacobian_t.append(terms)

    def __call__(self, x: np.ndarray) -> float:
        return self.evaluate(x)

    @abstractmethod
    def evaluate(self, x: np.ndarray) -> float:
        pass

    @abstractmethod
    def evaluate_jacobian(self, x: np.ndarray) -> np.ndarray:
        pass

    def _convert_input_to_model_tensor(
        self, x: np.ndarray, requires_grad: bool = True
    ) -> Tensor:
        """

        Args:
            x: x (np.ndarray): values of design variables a 1d array.
        """
        assert x.ndim == 1, "values of design should be 1d array"
        X = pd.DataFrame(
            x.reshape(len(x.flatten()) // self.n_vars, self.n_vars), columns=self.vars
        )
        X = self.model.get_model_matrix(X)
        return torch.tensor(X.values, requires_grad=requires_grad, **tkwargs)

    def _model_jacobian_t(self, x: np.ndarray) -> np.ndarray:
        """Computes the transpose of the model jacobian for each experiment in input x."""
        X = pd.DataFrame(x.reshape(self.n_experiments, self.n_vars), columns=self.vars)
        jacobians = np.swapaxes(X.eval(self.terms_jacobian_t), 0, 2)
        return np.swapaxes(jacobians, 1, 2)


class DOptimality(Objective):
    """A class implementing the evaluation of logdet(X.T@X + delta) and its jacobian w.r.t. the inputs.
    The Jacobian can be divided into two parts, one for logdet(X.T@ + delta) w.r.t. X (there is a simple
    closed expression for this one) and one model dependent part for the jacobian of X.T@X
    w.r.t. the inputs. Because each row of X only depends on the inputs of one experiment
    the second part can be formulated in a simplified way. It is built up with n_experiment
    blocks of the same structure which is represended by the attribute jacobian_building_block.

    A nice derivation for the "first part" of the jacobian can be found [here](https://angms.science/doc/LA/logdet.pdf).
    The second part consists of the partial derivatives of the model terms with
    respect to the inputs. We denote the value of the i-th model term from the j-th experiment
    with y_ij and the i-th input value of the j-th experiment with x_ij. N stands for the number
    of model terms, n for the number of input terms and M for the number of experiments.
    Here, we only consider models up to second order, but the computation can easily be extended
    for higher-ordermodels.

    To do the computation in the most basic way, we could compute the partial derivative of every
    single model term and experiment with respect to every single input and experiment. We could write
    this in one large matrix and multiply the first part of the gradient as a long vector from the right
    side.
    But because of the structure of the domain we can do the same computation with much smaller
    matrices:
    First, we write the first part of the jacobian as the matrix (df/dy_ij)_ij where i goes from 1 to N
    and j goes from 1 to M.

    Second, we compute a rank 3 tensor (K_kij)_kij. k goes from 1 to M, i from 1 to n and j from 1 to N.
    For each k (K_kij)_ij contains the partial derivatives (dy_jk/dx_ik)_ij. Note that the values of the
    entries of (dy_jk/dx_ik)_ij only depend on the input values of the k-th experiment. The function
    default_jacobian_building_block implements the computation of these matrices/"building blocks".

    Then, we notice that the model term values of the j-th experiment only depend on the input values of
    the j-th experiment. Thus, to compute the partial derivative df/dx_ik we only have to compute the euclidian
    scalar product of (K_kij)_j and (df/dy_jk)_j. The way how we built the two parts of the jacobian allows us
    to compute this scalar product in a vectorized way for all x_ik at once, see also JacobianForLogDet.jacobian.
    """

    def __init__(
        self,
        domain: Domain,
        model: Formula,
        n_experiments: int,
        delta: float = 1e-7,
    ) -> None:
        super().__init__(
            domain=domain, model=model, n_experiments=n_experiments, delta=delta
        )

    def evaluate(self, x: np.ndarray) -> float:
        """Computes the minus one times the sum of the log of the eigenvalues of X.T @ X + delta.
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            -log(det(X.T@X+delta))

        """
        X = self._convert_input_to_model_tensor(x, requires_grad=False)
        return float(
            -1
            * torch.logdet(
                X.detach().T @ X.detach() + self.delta * torch.eye(self.n_model_terms)
            )
        )

    def evaluate_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Computes the jacobian of minus one times the log of the determinant of X.T @ X + delta.
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            The jacobian of -log(det(X.T@X+delta)) as numpy array
        """
        # get model matrix X
        X = self._convert_input_to_model_tensor(x, requires_grad=True)

        # first part of jacobian
        torch.logdet(X.T @ X + self.delta * torch.eye(self.n_model_terms)).backward()
        J1 = -1 * X.grad.detach().numpy()  # type: ignore
        J1 = np.repeat(J1, self.n_vars, axis=0).reshape(
            self.n_experiments, self.n_vars, self.n_model_terms
        )

        # second part of jacobian
        J2 = self._model_jacobian_t(x)

        # combine both parts
        J = J1 * J2
        J = np.sum(J, axis=-1)

        return J.flatten()


class AOptimality(Objective):
    """A class implementing the evaluation of tr((X.T@X + delta)^-1) and its jacobian w.r.t. the inputs.
    The jacobian evaluation is done analogously to DOptimality with the first part of the jacobian being
    the jacobian of tr((X.T@X + delta)^-1) instead of logdet(X.T@X + delta).
    """

    def evaluate(self, x: np.ndarray) -> float:
        """Computes the trace of the inverse of X.T @ X + delta.
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            tr((X.T@X+delta)^-1)

        """
        X = self._convert_input_to_model_tensor(x, requires_grad=False)
        return float(
            torch.trace(
                torch.linalg.inv(
                    X.detach().T @ X.detach()
                    + self.delta * torch.eye(self.n_model_terms)
                )
            )
        )

    def evaluate_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Computes the jacobian of the trace of the inverse of X.T @ X + delta.
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            The jacobian of tr((X.T@X+delta)^-1) as numpy array
        """
        # get model matrix X
        X = self._convert_input_to_model_tensor(x, requires_grad=True)

        # first part of jacobian
        torch.trace(
            torch.linalg.inv(X.T @ X + self.delta * torch.eye(self.n_model_terms))
        ).backward()
        J1 = X.grad.detach().numpy()  # type: ignore
        J1 = np.repeat(J1, self.n_vars, axis=0).reshape(
            self.n_experiments, self.n_vars, self.n_model_terms
        )

        # second part of jacobian
        J2 = self._model_jacobian_t(x)

        # combine both parts
        J = J1 * J2
        J = np.sum(J, axis=-1)

        return J.flatten()


class GOptimality(Objective):
    """A class implementing the evaluation of max(diag(H)) and its jacobian w.r.t. the inputs where
    H = X @ (X.T@X + delta)^-1 @ X.T is the (regularized) hat matrix. The jacobian evaluation is done analogously
    to DOptimality with the first part of the jacobian being the jacobian of max(diag(H)) instead of
    logdet(X.T@X + delta).
    """

    def evaluate(self, x: np.ndarray) -> float:
        """Computes the maximum diagonal entry of H = X @ (X.T@X + delta)^-1 @ X.T .
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            max(diag(H))

        """
        X = self._convert_input_to_model_tensor(x, requires_grad=False)
        H = (
            X.detach()
            @ torch.linalg.inv(
                X.detach().T @ X.detach() + self.delta * torch.eye(self.n_model_terms)
            )
            @ X.detach().T
        )
        return float(torch.max(torch.diag(H)))

    def evaluate_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Computes the jacobian of the maximum diagonal element of H = X @ (X.T @ X + delta)^-1 @ X.T.
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            The jacobian of max(diag(H)) as numpy array
        """
        # get model matrix X
        X = self._convert_input_to_model_tensor(x, requires_grad=True)

        # first part of jacobian
        torch.max(
            torch.diag(
                X
                @ torch.linalg.inv(X.T @ X + self.delta * torch.eye(self.n_model_terms))
                @ X.T
            )
        ).backward()
        J1 = X.grad.detach().numpy()  # type: ignore
        J1 = np.repeat(J1, self.n_vars, axis=0).reshape(
            self.n_experiments, self.n_vars, self.n_model_terms
        )

        # second part of jacobian
        J2 = self._model_jacobian_t(x)

        # combine both parts
        J = J1 * J2
        J = np.sum(J, axis=-1)

        return J.flatten()


class EOptimality(Objective):
    """A class implementing the evaluation of minus one times the minimum eigenvalue of (X.T @ X + delta)
    and its jacobian w.r.t. the inputs. The jacobian evaluation is done analogously to DOptimality with the
    first part of the jacobian being the jacobian of the smallest eigenvalue of (X.T @ X + delta) instead of
    logdet(X.T@X + delta).
    """

    def evaluate(self, x: np.ndarray) -> float:
        """Computes minus one times the minimum eigenvalue of (X.T@X + delta).
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            min(eigvals(X.T @ X + delta))
        """
        X = self._convert_input_to_model_tensor(x, requires_grad=False)
        return -1 * float(
            torch.min(
                torch.linalg.eigvalsh(
                    X.detach().T @ X.detach()
                    + self.delta * torch.eye(self.n_model_terms)
                )
            )
        )

    def evaluate_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Computes the jacobian of minus one times the minimum eigenvalue of (X.T @ X + delta).
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            The jacobian of -1 * min(eigvals(X.T @ X + delta)) as numpy array
        """
        # get model matrix X
        X = self._convert_input_to_model_tensor(x, requires_grad=True)

        # first part of jacobian
        torch.min(
            torch.linalg.eigvalsh(X.T @ X + self.delta * torch.eye(self.n_model_terms))
        ).backward()
        J1 = -1 * X.grad.detach().numpy()  # type: ignore
        J1 = np.repeat(J1, self.n_vars, axis=0).reshape(
            self.n_experiments, self.n_vars, self.n_model_terms
        )

        # second part of jacobian
        J2 = self._model_jacobian_t(x)

        # combine both parts
        J = J1 * J2
        J = np.sum(J, axis=-1)

        return J.flatten()


class KOptimality(Objective):
    """A class implementing the evaluation of the condition number of (X.T @ X + delta)
    and its jacobian w.r.t. the inputs. The jacobian evaluation is done analogously to
    DOptimality with the first part of the jacobian being the jacobian of condition number
    of (X.T @ X + delta) instead of logdet(X.T@X + delta).
    """

    def evaluate(self, x: np.ndarray) -> float:
        """Computes condition number of (X.T@X + delta).
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            cond(X.T @ X + delta)
        """
        X = self._convert_input_to_model_tensor(x, requires_grad=False)
        return float(
            torch.linalg.cond(
                X.detach().T @ X.detach() + self.delta * torch.eye(self.n_model_terms)
            )
        )

    def evaluate_jacobian(self, x: np.ndarray) -> np.ndarray:
        """Computes the jacobian of the condition number of (X.T @ X + delta).
        Where X is the model matrix corresponding to x.

        Args:
            x (np.ndarray): values of design variables a 1d array.

        Returns:
            The jacobian of cond(X.T @ X + delta) as numpy array
        """
        # get model matrix X
        X = self._convert_input_to_model_tensor(x, requires_grad=True)

        # first part of jacobian
        torch.linalg.cond(
            X.T @ X + self.delta * torch.eye(self.n_model_terms)
        ).backward()
        J1 = X.grad.detach().numpy()  # type: ignore
        J1 = np.repeat(J1, self.n_vars, axis=0).reshape(
            self.n_experiments, self.n_vars, self.n_model_terms
        )

        # second part of jacobian
        J2 = self._model_jacobian_t(x)

        # combine both parts
        J = J1 * J2
        J = np.sum(J, axis=-1)

        return J.flatten()


class SpaceFilling(Objective):
    def evaluate(self, x: np.ndarray) -> float:
        X = self._convert_input_to_tensor(x, requires_grad=False)
        return float(
            -torch.sum(torch.sort(torch.pdist(X.detach()))[0][: self.n_experiments])
        )

    def evaluate_jacobian(self, x: np.ndarray) -> float:
        X = self._convert_input_to_tensor(x, requires_grad=True)
        torch.sum(torch.sort(torch.pdist(X))[0][: self.n_experiments]).backward()

        return -X.grad.detach().numpy().flatten()  # type: ignore

    def _convert_input_to_tensor(
        self, x: np.ndarray, requires_grad: bool = True
    ) -> Tensor:
        X = pd.DataFrame(
            x.reshape(len(x.flatten()) // self.n_vars, self.n_vars), columns=self.vars
        )
        return torch.tensor(X.values, requires_grad=requires_grad, **tkwargs)


def get_objective_class(objective: OptimalityCriterionEnum) -> Type:
    objective = OptimalityCriterionEnum(objective)

    if objective == OptimalityCriterionEnum.D_OPTIMALITY:
        return DOptimality
    elif objective == OptimalityCriterionEnum.A_OPTIMALITY:
        return AOptimality
    elif objective == OptimalityCriterionEnum.G_OPTIMALITY:
        return GOptimality
    elif objective == OptimalityCriterionEnum.E_OPTIMALITY:
        return EOptimality
    elif objective == OptimalityCriterionEnum.K_OPTIMALITY:
        return KOptimality
    elif objective == OptimalityCriterionEnum.SPACE_FILLING:
        return SpaceFilling



from abc import abstractmethod
from typing import Union

import pandas as pd

import bofire.data_models.strategies.stepwise.conditions as data_models
from bofire.data_models.domain.api import Domain


class Condition(object):
    @abstractmethod
    def evaluate(self, domain: Domain, experiments: pd.DataFrame) -> bool:
        pass


class NumberOfExperimentsCondition(Condition):
    def __init__(self, data_model: data_models.NumberOfExperimentsCondition):
        self.n_experiments = data_model.n_experiments

    def evaluate(self, domain: Domain, experiments: Union[pd.DataFrame, None]) -> bool:
        if experiments is None:
            n_experiments = 0
        else:
            n_experiments = len(
                domain.outputs.preprocess_experiments_all_valid_outputs(experiments)
            )
        return n_experiments <= self.n_experiments


class AlwaysTrueCondition(Condition):
    def __init__(self, data_model: data_models.AlwaysTrueCondition):
        pass

    def evaluate(self, domain: Domain, experiments: pd.DataFrame) -> bool:
        return True


class CombiCondition(Condition):
    def __init__(self, data_model: data_models.CombiCondition) -> None:
        self.conditions = [map(c) for c in data_model.conditions]  # type: ignore
        self.n_required_conditions = data_model.n_required_conditions

    def evaluate(self, domain: Domain, experiments: pd.DataFrame) -> bool:
        n_matched_conditions = 0
        for c in self.conditions:
            if c.evaluate(domain, experiments):
                n_matched_conditions += 1
        if n_matched_conditions >= self.n_required_conditions:
            return True
        return False


CONDITION_MAP = {
    data_models.CombiCondition: CombiCondition,
    data_models.NumberOfExperimentsCondition: NumberOfExperimentsCondition,
    data_models.AlwaysTrueCondition: AlwaysTrueCondition,
}


def map(
    data_model: Union[
        data_models.CombiCondition,
        data_models.NumberOfExperimentsCondition,
        data_models.AlwaysTrueCondition,
    ],
) -> Union[CombiCondition, NumberOfExperimentsCondition, AlwaysTrueCondition]:
    return CONDITION_MAP[data_model.__class__](data_model)



import pytest

import bofire.data_models.strategies.api as data_models
import bofire.strategies.stepwise.conditions as conditions
from bofire.benchmarks.single import Himmelblau


def test_RequiredExperimentsCondition():
    benchmark = Himmelblau()
    experiments = benchmark.f(benchmark.domain.inputs.sample(3), return_complete=True)
    data_model = data_models.NumberOfExperimentsCondition(n_experiments=3)
    condition = conditions.map(data_model=data_model)
    assert condition.evaluate(benchmark.domain, experiments=experiments) is True
    experiments = benchmark.f(benchmark.domain.inputs.sample(10), return_complete=True)
    assert condition.evaluate(benchmark.domain, experiments=experiments) is False


def test_RequiredExperimentsCondition_no_experiments():
    data_model = data_models.NumberOfExperimentsCondition(n_experiments=3)
    condition = conditions.map(data_model=data_model)
    assert condition.evaluate(Himmelblau().domain, experiments=None) is True


def test_AlwaysTrueCondition():
    benchmark = Himmelblau()
    experiments = benchmark.f(benchmark.domain.inputs.sample(3), return_complete=True)
    data_model = data_models.AlwaysTrueCondition()
    condition = conditions.map(data_model=data_model)
    assert condition.evaluate(benchmark.domain, experiments=experiments) is True


def test_CombiCondition_invalid():
    with pytest.raises(
        ValueError,
        match="Number of required conditions larger than number of conditions.",
    ):
        data_models.CombiCondition(
            conditions=[
                data_models.NumberOfExperimentsCondition(n_experiments=2),
                data_models.NumberOfExperimentsCondition(n_experiments=3),
            ],
            n_required_conditions=3,
        )


@pytest.mark.parametrize(
    "n_required, n_experiments, expected", [(1, 10, True), (2, 1, True)]
)
def test_CombiCondition(n_required, n_experiments, expected):
    benchmark = Himmelblau()
    experiments = benchmark.f(
        benchmark.domain.inputs.sample(n_experiments), return_complete=True
    )
    data_model = data_models.CombiCondition(
        conditions=[
            data_models.NumberOfExperimentsCondition(n_experiments=2),
            data_models.NumberOfExperimentsCondition(n_experiments=12),
        ],
        n_required_conditions=n_required,
    )
    condition = conditions.map(data_model=data_model)
    assert condition.evaluate(benchmark.domain, experiments=experiments) is expected



import numpy as np
import pytest

from bofire.benchmarks.single import (
    Ackley,
    Branin,
    Branin30,
    DiscreteHimmelblau,
    Hartmann,
    Himmelblau,
    _CategoricalDiscreteHimmelblau,
)


def test_hartmann():
    with pytest.raises(ValueError):
        Hartmann(8)
    h = Hartmann(dim=6, allowed_k=3)
    assert h.dim == 6
    assert h.domain.constraints[0].max_count == 3
    with pytest.raises(ValueError):
        h.get_optima()
    h = Hartmann(dim=6, allowed_k=None)
    assert len(h.domain.constraints) == 0


@pytest.mark.parametrize(
    "cls_benchmark, return_complete, kwargs",
    [
        (Himmelblau, False, {}),
        (DiscreteHimmelblau, False, {}),
        (DiscreteHimmelblau, True, {}),
        (_CategoricalDiscreteHimmelblau, False, {}),
        (_CategoricalDiscreteHimmelblau, True, {}),
        (Ackley, False, {}),
        (Himmelblau, True, {}),
        (Ackley, True, {}),
        (Hartmann, True, {}),
        (Hartmann, False, {}),
        (Branin, True, {}),
        (Branin, False, {}),
        (Branin30, True, {}),
        (Branin30, False, {})
        # TO DO: Implement feature that tests Ackley for categorical and descriptive inputs.
        # (Ackley, {"categorical": True}),
        # (Ackley, {"descriptor": True}),
    ],
)
def test_single_objective_benchmarks(cls_benchmark, return_complete, kwargs):
    """Test function for single objective benchmark functions.

    Args:
        cls_benchmark (Benchmark function class): Benchmark function that is supposed to be tested.
        kwargs ({"dim": , "k":}): Optinal arguments for benchmark functions that require additional arguments. Ackley can handle categerical and descriptive inputs.
    """
    benchmark_function = cls_benchmark(**kwargs)
    benchmark_function_name = benchmark_function.__class__.__name__

    # Check for correct dimensions
    n_samples = 1000
    X_samples = benchmark_function.domain.inputs.sample(n=n_samples)
    # Calculating corresponding y values
    Y = benchmark_function.f(X_samples, return_complete=return_complete)
    # Check, whether shape of output dataframe matches the expected shape.
    expected_output_variables = len(benchmark_function.domain.outputs) * 2

    if return_complete:
        assert Y.shape == (
            n_samples,
            len(benchmark_function.domain.experiment_column_names),
        ), (
            "The shape of the output dataframe of "
            + benchmark_function_name
            + " does not match the expected shape."
        )
    else:
        assert Y.shape == (n_samples, expected_output_variables), (
            "The shape of the output dataframe of "
            + benchmark_function_name
            + " does not match the expected shape."
        )

    # Check for correct optima
    # Retrieve optima from benchmark function
    try:
        optima = benchmark_function.get_optima()
        # Retrieve variable names
        x_keys = benchmark_function.domain.inputs.get_keys()
        y_keys = benchmark_function.domain.outputs.get_keys()
        px = optima[x_keys]
        py = optima[y_keys]
        Y = benchmark_function.f(px, return_complete=return_complete)
        y = Y[y_keys]
        # Check whether the optimum y value at position px equals the expected value at px with tolerance 'atol'.
        assert np.allclose(y, py, atol=1e-04), (
            "Expected optimum value of "
            + benchmark_function_name
            + " does not match calculated optimum value or is out of the tolerance radius."
        )
    except NotImplementedError:
        pass



import numpy as np
from pandas.testing import assert_frame_equal

import bofire.surrogates.api as surrogates
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import ContinuousInput, ContinuousOutput
from bofire.data_models.kernels.api import PolynomialKernel
from bofire.data_models.surrogates.api import BotorchSurrogates, PolynomialSurrogate


def test_polynomial_surrogate():
    N_EXPERIMENTS = 10

    inputs = Inputs(
        features=[
            ContinuousInput(key="a", bounds=(0, 40)),
            ContinuousInput(key="b", bounds=(20, 60)),
        ]
    )
    outputs = Outputs(features=[ContinuousOutput(key="c")])

    experiments = inputs.sample(N_EXPERIMENTS)
    experiments["c"] = (
        experiments["a"] * 2.2
        + experiments["b"] * -0.05
        + experiments["b"]
        + np.random.normal(loc=0, scale=5, size=N_EXPERIMENTS)
    )
    experiments["valid_c"] = 1

    surrogate_data = PolynomialSurrogate.from_power(
        power=2, inputs=inputs, outputs=outputs
    )
    surrogate = surrogates.map(surrogate_data)

    assert isinstance(surrogate, surrogates.SingleTaskGPSurrogate)
    assert isinstance(surrogate.kernel, PolynomialKernel)

    # check dump
    surrogate.fit(experiments=experiments)
    preds = surrogate.predict(experiments)
    dump = surrogate.dumps()
    surrogate.loads(dump)
    preds2 = surrogate.predict(experiments)
    assert_frame_equal(preds, preds2)


def test_can_define_botorch_surrogate():
    inputs = Inputs(
        features=[
            ContinuousInput(key="a", bounds=(0, 40)),
            ContinuousInput(key="b", bounds=(20, 80)),
        ]
    )
    outputs = [ContinuousOutput(key="c"), ContinuousOutput(key="d")]
    (
        BotorchSurrogates(
            surrogates=[
                PolynomialSurrogate(
                    inputs=inputs, outputs=Outputs(features=[outputs[0]])
                ),
                PolynomialSurrogate(
                    inputs=inputs, outputs=Outputs(features=[outputs[1]])
                ),
            ]
        ),
    )



from functools import partial
from typing import Union

from bofire.data_models.priors.gamma import GammaPrior
from bofire.data_models.priors.normal import NormalPrior
from bofire.data_models.priors.prior import Prior

AbstractPrior = Prior

AnyPrior = Union[
    GammaPrior,
    NormalPrior,
]

# default priors of interest
# botorch defaults
BOTORCH_LENGTHCALE_PRIOR = partial(GammaPrior, concentration=3.0, rate=6.0)
BOTORCH_NOISE_PRIOR = partial(GammaPrior, concentration=1.1, rate=0.05)
BOTORCH_SCALE_PRIOR = partial(GammaPrior, concentration=2.0, rate=0.15)

# mbo priors
# By default BoTorch places a highly informative prior on the kernel lengthscales,
# which easily leads to overfitting. Here we set a broader prior distribution for the
# lengthscale. The priors for the noise and signal variance are set more tightly.
MBO_LENGTHCALE_PRIOR = partial(GammaPrior, concentration=2.0, rate=0.2)
MBO_NOISE_PRIOR = partial(GammaPrior, concentration=2.0, rate=4.0)
MBO_OUTPUTSCALE_PRIOR = partial(GammaPrior, concentration=2.0, rate=4.0)



from typing import Union

import pytest
from pydantic import BaseModel

import bofire.data_models.unions as unions


class A(BaseModel):
    pass


class B(BaseModel):
    pass


class C(BaseModel):
    pass


ABC = Union[A, B, C]
AB = Union[A, B]
AC = Union[A, C]


@pytest.mark.parametrize(
    "union, expected",
    [
        (ABC, {A, B, C}),
        (AB, {A, B}),
        (AC, {A, C}),
    ],
)
def test_union_to_list_should_process_union(union, expected):
    assert set(unions.to_list(union)) == expected


@pytest.mark.parametrize(
    "single",
    [
        A,
        B,
        C,
    ],
)
def test_union_to_list_should_process_single_model(single):
    assert unions.to_list(single) == [single]



import importlib

import numpy as np
import pytest

from bofire.utils.cheminformatics import (  # smiles2bag_of_characters,
    smiles2fingerprints,
    smiles2fragments,
    smiles2fragments_fingerprints,
    smiles2mol,
    smiles2mordred,
)

RDKIT_AVAILABLE = importlib.util.find_spec("rdkit") is not None


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_smiles2mol():
    # invalid
    with pytest.raises(ValueError):
        smiles2mol("abcd")
    # valid
    smiles2mol("CC(=O)Oc1ccccc1C(=O)O")


smiles = [
    "CC(=O)Oc1ccccc1C(=O)O",
    "c1ccccc1",
    "[CH3][CH2][OH]",
    "N[C@](C)(F)C(=O)O",
]


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_smiles2fingerprints():
    values = np.array(
        [
            [
                1,
                1,
                1,
                1,
                1,
                1,
                0,
                1,
                1,
                1,
                1,
                1,
                0,
                1,
                0,
                1,
                1,
                1,
                1,
                0,
                0,
                0,
                1,
                1,
                0,
                1,
                0,
                1,
                1,
                1,
                0,
                0,
            ],
            [
                1,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ],
            [
                0,
                1,
                1,
                0,
                0,
                0,
                1,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                1,
                0,
            ],
            [
                0,
                1,
                0,
                1,
                0,
                1,
                0,
                1,
                1,
                0,
                1,
                0,
                0,
                1,
                0,
                0,
                0,
                0,
                1,
                1,
                0,
                0,
                0,
                1,
                0,
                1,
                0,
                0,
                0,
                1,
                0,
                0,
            ],
        ]
    )
    desc = smiles2fingerprints(smiles=smiles, n_bits=32)
    assert desc.shape[0] == 4
    assert desc.shape[1] == 32
    np.testing.assert_array_equal(desc, values)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_smiles2fragments():
    values = np.array([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]])

    desc = smiles2fragments(
        smiles=smiles, fragments_list=["fr_unbrch_alkane", "fr_thiocyan"]
    )
    assert desc.shape[0] == 4
    assert desc.shape[1] == 2
    np.testing.assert_array_equal(desc, values)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_smiles2fragments_fingerprints():
    values = np.array(
        [
            [
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                0.0,
                1.0,
                1.0,
                1.0,
                1.0,
                1.0,
                0.0,
                1.0,
                0.0,
                1.0,
                1.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                1.0,
                1.0,
                0.0,
                1.0,
                0.0,
                1.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ],
            [
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ],
            [
                0.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
            ],
            [
                0.0,
                1.0,
                0.0,
                1.0,
                0.0,
                1.0,
                0.0,
                1.0,
                1.0,
                0.0,
                1.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ],
        ]
    )

    desc = smiles2fragments_fingerprints(
        smiles=smiles, n_bits=32, fragments_list=["fr_unbrch_alkane", "fr_thiocyan"]
    )
    assert desc.shape[0] == 4
    assert desc.shape[1] == 32 + 2
    np.testing.assert_array_equal(desc, values)


@pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
def test_smiles2mordred():
    values = np.array(
        [
            [0.5963718820861676, 0.0],
            [-1.5, 0.0],
            [-0.28395061728395066, 1.0],
            [-8.34319526627219, 0.0],
        ]
    )

    desc = smiles2mordred(smiles=smiles, descriptors_list=["NssCH2", "ATSC2d"])
    assert desc.shape[0] == 4
    assert desc.shape[1] == 2
    np.testing.assert_array_almost_equal(desc, values)


# @pytest.mark.skipif(not RDKIT_AVAILABLE, reason="requires rdkit")
# def test_smiles2bag_of_characters():
#     smiles = ["CC(=O)Oc1ccccc1C(=O)O", "c1ccccc1"]
#     desc = smiles2bag_of_characters(smiles=smiles)
#     assert desc.shape[0] == 2



from functools import partial
from typing import Dict, Optional

import botorch
import pandas as pd
import torch
from botorch.fit import fit_gpytorch_mll
from botorch.models.transforms.input import ChainedInputTransform, OneHotToNumeric
from botorch.models.transforms.outcome import Standardize
from gpytorch.mlls import ExactMarginalLogLikelihood

import bofire.kernels.api as kernels
from bofire.data_models.enum import CategoricalEncodingEnum, OutputFilteringEnum
from bofire.data_models.surrogates.api import MixedSingleTaskGPSurrogate as DataModel
from bofire.surrogates.botorch import BotorchSurrogate
from bofire.surrogates.single_task_gp import get_scaler
from bofire.surrogates.trainable import TrainableSurrogate
from bofire.utils.torch_tools import tkwargs


class MixedSingleTaskGPSurrogate(BotorchSurrogate, TrainableSurrogate):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        self.continuous_kernel = data_model.continuous_kernel
        self.categorical_kernel = data_model.categorical_kernel
        self.scaler = data_model.scaler
        super().__init__(data_model=data_model, **kwargs)

    model: Optional[botorch.models.MixedSingleTaskGP] = None
    _output_filtering: OutputFilteringEnum = OutputFilteringEnum.ALL
    # features2idx: Optional[Dict] = None
    # non_numerical_features: Optional[List] = None
    training_specs: Dict = {}

    def _fit(self, X: pd.DataFrame, Y: pd.DataFrame):
        scaler = get_scaler(self.inputs, self.input_preprocessing_specs, self.scaler, X)
        transformed_X = self.inputs.transform(X, self.input_preprocessing_specs)

        tX, tY = torch.from_numpy(transformed_X.values).to(**tkwargs), torch.from_numpy(
            Y.values
        ).to(**tkwargs)

        features2idx, _ = self.inputs._get_transform_info(
            self.input_preprocessing_specs
        )
        non_numerical_features = [
            key
            for key, value in self.input_preprocessing_specs.items()
            if value != CategoricalEncodingEnum.DESCRIPTOR
        ]

        ord_dims = []
        for feat in self.inputs.get():
            if feat.key not in non_numerical_features:
                ord_dims += features2idx[feat.key]

        # these are the categorical dimesions after applying the OneHotToNumeric transform
        cat_dims = list(
            range(len(ord_dims), len(ord_dims) + len(non_numerical_features))
        )
        # these are the categorical features within the the OneHotToNumeric transform
        categorical_features = {
            features2idx[feat][0]: len(features2idx[feat])
            for feat in non_numerical_features
        }

        o2n = OneHotToNumeric(
            dim=tX.shape[1],
            categorical_features=categorical_features,
            transform_on_train=False,
        )
        tf = ChainedInputTransform(tf1=scaler, tf2=o2n) if scaler is not None else o2n

        # fit the model
        self.model = botorch.models.MixedSingleTaskGP(
            train_X=o2n.transform(tX),
            train_Y=tY,
            cat_dims=cat_dims,
            # cont_kernel_factory=self.continuous_kernel.to_gpytorch,
            cont_kernel_factory=partial(kernels.map, data_model=self.continuous_kernel),
            outcome_transform=Standardize(m=tY.shape[-1]),
            input_transform=tf,
        )
        mll = ExactMarginalLogLikelihood(self.model.likelihood, self.model)
        fit_gpytorch_mll(mll, options=self.training_specs)



import itertools
from typing import Annotated, List

from pydantic import Field, validator

from bofire.data_models.base import BaseModel
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.outlier_detection.outlier_detection import IterativeTrimming

AnyOutlierDetector = IterativeTrimming


class OutlierDetections(BaseModel):
    """ "List of Outlier detectors.

    Behaves similar to a outlier_detector."""

    detectors: Annotated[List[AnyOutlierDetector], Field(min_items=1)]

    @property
    def outputs(self) -> Outputs:
        return Outputs(
            features=list(
                itertools.chain.from_iterable(
                    [model.outputs.get() for model in self.detectors]  # type: ignore
                )
            )
        )

    @validator("detectors")
    def validate_detectors(cls, v, values):
        used_output_feature_keys = list(
            itertools.chain.from_iterable(
                [detector.outputs.get_keys() for detector in v]
            )
        )
        if len(set(used_output_feature_keys)) != len(used_output_feature_keys):
            raise ValueError("Output feature keys are not unique across detectors.")
        return v

    def _check_compability(self, inputs: Inputs, outputs: Outputs):
        # TODO: add sync option
        used_output_feature_keys = self.outputs.get_keys()
        if sorted(used_output_feature_keys) != sorted(outputs.get_keys()):  # type: ignore
            raise ValueError("Output features do not match.")
        used_feature_keys = []
        for i, model in enumerate(self.detectors):
            if len(model.inputs) > len(inputs):  # type: ignore
                raise ValueError(
                    f"Model with index {i} has more features than acceptable."
                )
            for feat in model.inputs:  # type: ignore
                try:
                    other_feat = inputs.get_by_key(feat.key)
                except KeyError:
                    raise ValueError(f"Feature {feat.key} not found.")
                # now compare the features
                # TODO: make more sohisticated comparisons based on the type
                # has to to be implemented in features, for the start
                # we go with __eq__
                if feat != other_feat:
                    raise ValueError(f"Features with key {feat.key} are incompatible.")
                if feat.key not in used_feature_keys:
                    used_feature_keys.append(feat.key)
        if len(used_feature_keys) != len(inputs):
            raise ValueError("Unused features are present.")



from typing import Literal

from bofire.data_models.kernels.kernel import Kernel


class CategoricalKernel(Kernel):
    pass


class HammondDistanceKernel(CategoricalKernel):
    type: Literal["HammondDistanceKernel"] = "HammondDistanceKernel"
    ard: bool = True



from typing import Dict, Optional, Union

import botorch
import pandas as pd
import torch
from botorch.fit import fit_gpytorch_mll
from botorch.models.transforms.input import InputStandardize, Normalize
from botorch.models.transforms.outcome import Standardize
from gpytorch.mlls import ExactMarginalLogLikelihood

import bofire.kernels.api as kernels
import bofire.priors.api as priors
from bofire.data_models.domain.api import Inputs
from bofire.data_models.enum import CategoricalEncodingEnum, OutputFilteringEnum
from bofire.data_models.features.api import TInputTransformSpecs
from bofire.data_models.molfeatures.api import MolFeatures
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate as DataModel
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.surrogates.botorch import BotorchSurrogate
from bofire.surrogates.trainable import TrainableSurrogate
from bofire.utils.torch_tools import tkwargs


def get_scaler(
    inputs: Inputs,
    input_preprocessing_specs: TInputTransformSpecs,
    scaler: ScalerEnum,
    X: pd.DataFrame,
) -> Union[InputStandardize, Normalize, None]:
    """Returns the instanitated scaler object for a set of input features and
    input_preprocessing_specs.


    Args:
        inputs (Inputs): Input features.
        input_preprocessing_specs (TInputTransformSpecs): Dictionary how to treat
            the categoricals and/or molecules.
        scaler (ScalerEnum): Enum indicating the scaler of interest.
        X (pd.DataFrame): The dataset of interest.

    Returns:
        Union[InputStandardize, Normalize]: The instantiated scaler class
    """
    if scaler != ScalerEnum.IDENTITY:
        features2idx, _ = inputs._get_transform_info(input_preprocessing_specs)

        d = 0
        for indices in features2idx.values():
            d += len(indices)

        non_numerical_features = [
            key
            for key, value in input_preprocessing_specs.items()
            if value != CategoricalEncodingEnum.DESCRIPTOR
            and not isinstance(value, MolFeatures)
        ]

        ord_dims = []
        for feat in inputs.get():
            if feat.key not in non_numerical_features:
                ord_dims += features2idx[feat.key]

        if scaler == ScalerEnum.NORMALIZE:
            lower, upper = inputs.get_bounds(
                specs=input_preprocessing_specs, experiments=X
            )
            scaler_transform = Normalize(
                d=d,
                bounds=torch.tensor([lower, upper]).to(**tkwargs),
                indices=ord_dims,
                batch_shape=torch.Size(),
            )
        elif scaler == ScalerEnum.STANDARDIZE:
            scaler_transform = InputStandardize(
                d=d,
                indices=ord_dims,
                batch_shape=torch.Size(),
            )
        else:
            raise ValueError("Scaler enum not known.")
        return scaler_transform
    else:
        return None


class SingleTaskGPSurrogate(BotorchSurrogate, TrainableSurrogate):
    def __init__(
        self,
        data_model: DataModel,
        **kwargs,
    ):
        self.kernel = data_model.kernel
        self.scaler = data_model.scaler
        self.noise_prior = data_model.noise_prior
        super().__init__(data_model=data_model, **kwargs)

    model: Optional[botorch.models.SingleTaskGP] = None
    _output_filtering: OutputFilteringEnum = OutputFilteringEnum.ALL
    training_specs: Dict = {}

    def _fit(self, X: pd.DataFrame, Y: pd.DataFrame):
        scaler = get_scaler(self.inputs, self.input_preprocessing_specs, self.scaler, X)
        transformed_X = self.inputs.transform(X, self.input_preprocessing_specs)

        tX, tY = torch.from_numpy(transformed_X.values).to(**tkwargs), torch.from_numpy(
            Y.values
        ).to(**tkwargs)

        self.model = botorch.models.SingleTaskGP(  # type: ignore
            train_X=tX,
            train_Y=tY,
            covar_module=kernels.map(
                self.kernel,
                batch_shape=torch.Size(),
                active_dims=list(range(tX.shape[1])),
                ard_num_dims=1,  # this keyword is ingored
            ),
            outcome_transform=Standardize(m=tY.shape[-1]),
            input_transform=scaler,
        )

        self.model.likelihood.noise_covar.noise_prior = priors.map(self.noise_prior)  # type: ignore

        mll = ExactMarginalLogLikelihood(self.model.likelihood, self.model)
        fit_gpytorch_mll(mll, options=self.training_specs, max_attempts=10)



from typing import List

import pytest
from pydantic.error_wrappers import ValidationError

from bofire.data_models.priors.api import GammaPrior, NormalPrior


def get_invalids(valid: dict) -> List[dict]:
    return [
        {k: v for k, v in valid.items() if k != k_}
        for k_ in valid.keys()
        if k_ != "type"
    ]


VALID_GAMMA_PRIOR_SPEC = {"type": "GammaPrior", "concentration": 2.0, "rate": 0.2}

VALID_NORMAL_PRIOR_SPEC = {"type": "NormalPrior", "loc": 2.0, "scale": 0.2}

PRIOR_SPECS = {
    GammaPrior: {
        "valids": [
            VALID_GAMMA_PRIOR_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_GAMMA_PRIOR_SPEC),
            *[
                {
                    **VALID_GAMMA_PRIOR_SPEC,
                    "concentration": concentration,
                    "rate": rate,
                }
                for rate in [-1.0, 0]
                for concentration in [-5, 6]
            ],
        ],
    },
    NormalPrior: {
        "valids": [
            VALID_NORMAL_PRIOR_SPEC,
        ],
        "invalids": [
            *get_invalids(VALID_NORMAL_PRIOR_SPEC),
            *[
                {
                    **VALID_NORMAL_PRIOR_SPEC,
                    "loc": 0,
                    "scale": scale,
                }
                for scale in [-1.0, 0]
            ],
        ],
    },
}


@pytest.mark.parametrize(
    "cls, spec",
    [(cls, valid) for cls, data in PRIOR_SPECS.items() for valid in data["valids"]],
)
def test_valid_prior_specs(cls, spec):
    res = cls(**spec)
    assert isinstance(res, cls)
    assert isinstance(res.__str__(), str)


@pytest.mark.parametrize(
    "cls, spec",
    [
        (cls, invalid)
        for cls, data in PRIOR_SPECS.items()
        for invalid in data["invalids"]
    ],
)
def test_invalid_prior_specs(cls, spec):
    with pytest.raises((ValueError, TypeError, KeyError, ValidationError)):
        _ = cls(**spec)



from abc import abstractmethod
from typing import List, Literal, Optional

import pandas as pd
from pydantic import Field, validator
from typing_extensions import Annotated

import bofire.data_models.molfeatures.names as names
from bofire.data_models.base import BaseModel
from bofire.utils.cheminformatics import (  # smiles2bag_of_characters,
    smiles2fingerprints,
    smiles2fragments,
    smiles2fragments_fingerprints,
    smiles2mordred,
)


class MolFeatures(BaseModel):
    """Base class for all molecular features"""

    type: str

    @abstractmethod
    def get_descriptor_names(self) -> List[str]:
        pass

    @abstractmethod
    def get_descriptor_values(self, values: pd.Series) -> pd.DataFrame:
        pass


class Fingerprints(MolFeatures):
    type: Literal["Fingerprints"] = "Fingerprints"
    bond_radius: int = 5
    n_bits: int = 2048

    def get_descriptor_names(self) -> List[str]:
        return [f"fingerprint_{i}" for i in range(self.n_bits)]

    def get_descriptor_values(self, values: pd.Series) -> pd.DataFrame:
        return pd.DataFrame(
            data=smiles2fingerprints(
                values.to_list(), bond_radius=self.bond_radius, n_bits=self.n_bits
            ).astype(float),
            columns=self.get_descriptor_names(),
            index=values.index,
        )


class Fragments(MolFeatures):
    type: Literal["Fragments"] = "Fragments"
    fragments: Optional[List[str]] = None

    @validator(
        "fragments",
    )
    def validate_fragments(cls, fragments):
        """validates that fragments have unique names

        Args:
            categories (List[str]): List of fragment names

        Raises:
            ValueError: when fragments have non-unique names

        Returns:
            List[str]: List of the fragments
        """
        if fragments is not None:
            if len(fragments) != len(set(fragments)):
                raise ValueError("Fragments must be unique")

            if not all(user_fragment in names.fragments for user_fragment in fragments):
                raise ValueError(
                    "Not all provided fragments were not found in the RDKit list"
                )

        return fragments

    def get_descriptor_names(self) -> List[str]:
        return self.fragments if self.fragments is not None else names.fragments

    def get_descriptor_values(self, values: pd.Series) -> pd.DataFrame:
        return pd.DataFrame(
            data=smiles2fragments(values.to_list(), self.get_descriptor_names()),
            columns=self.get_descriptor_names(),
            index=values.index,
        )


class FingerprintsFragments(Fingerprints, Fragments):
    type: Literal["FingerprintsFragments"] = "FingerprintsFragments"

    def get_descriptor_names(self) -> List[str]:
        fingerprints_list = [f"fingerprint_{i}" for i in range(self.n_bits)]
        fragments_list = (
            self.fragments if self.fragments is not None else names.fragments
        )

        fingerprints_fragment_list = fingerprints_list + fragments_list

        return fingerprints_fragment_list

    def get_descriptor_values(self, values: pd.Series) -> pd.DataFrame:
        fragments_list = (
            self.fragments if self.fragments is not None else names.fragments
        )

        return pd.DataFrame(
            data=smiles2fragments_fingerprints(
                values.to_list(),
                bond_radius=self.bond_radius,
                n_bits=self.n_bits,
                fragments_list=fragments_list,
            ),
            columns=self.get_descriptor_names(),
            index=values.index,
        )


class MordredDescriptors(MolFeatures):
    type: Literal["MordredDescriptors"] = "MordredDescriptors"
    descriptors: Annotated[List[str], Field(min_items=1)]

    @validator(
        "descriptors",
    )
    def validate_descriptors(cls, descriptors):
        """validates that descriptors have unique names

        Args:
            descriptors (List[str]): List of descriptor names

        Raises:
            ValueError: when descriptors have non-unique names

        Returns:
            List[str]: List of the descriptors
        """
        if len(descriptors) != len(set(descriptors)):
            raise ValueError("descriptors must be unique")

        if not all(desc in names.mordred for desc in descriptors):
            raise ValueError(
                "Not all provided descriptors were not found in the Mordred list"
            )

        return descriptors

    def get_descriptor_names(self) -> List[str]:
        return self.descriptors

    def get_descriptor_values(self, values: pd.Series) -> pd.DataFrame:
        return pd.DataFrame(
            data=smiles2mordred(values.to_list(), self.descriptors),
            columns=self.descriptors,
            index=values.index,
        )



from abc import abstractmethod
from typing import ClassVar, Dict, List, Optional, Tuple, Union

import pandas as pd
from pydantic import Field
from typing_extensions import Annotated

from bofire.data_models.base import BaseModel
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.molfeatures.api import AnyMolFeatures
from bofire.data_models.surrogates.scaler import ScalerEnum

TTransform = Union[CategoricalEncodingEnum, ScalerEnum]


class Feature(BaseModel):
    """The base class for all features."""

    type: str
    key: str
    order_id: ClassVar[int] = -1

    def __lt__(self, other) -> bool:
        """
        Method to compare two models to get them in the desired order.
        Return True if other is larger than self, else False. (see FEATURE_ORDER)

        Args:
            other: The other class to compare to self

        Returns:
            bool: True if the other class is larger than self, else False
        """
        order_self = self.order_id
        order_other = other.order_id
        if order_self == order_other:
            return self.key < other.key
        else:
            return order_self < order_other


class Input(Feature):
    """Base class for all input features."""

    @staticmethod
    @abstractmethod
    def valid_transform_types() -> List[Union[CategoricalEncodingEnum, AnyMolFeatures]]:
        pass

    @abstractmethod
    def is_fixed(self) -> bool:
        """Indicates if a variable is set to a fixed value.

        Returns:
            bool: True if fixed, els False.
        """
        pass

    @abstractmethod
    def fixed_value(
        self, transform_type: Optional[TTransform] = None
    ) -> Union[None, List[str], List[float]]:
        """Method to return the fixed value in case of a fixed feature.

        Returns:
            Union[None,str,float]: None in case the feature is not fixed, else the fixed value.
        """
        pass

    @abstractmethod
    def validate_experimental(
        self, values: pd.Series, strict: bool = False
    ) -> pd.Series:
        """Abstract method to validate the experimental dataFrame

        Args:
            values (pd.Series): A dataFrame with experiments
            strict (bool, optional): Boolean to distinguish if the occurence of fixed features in the dataset should be considered or not. Defaults to False.

        Returns:
            pd.Series: The passed dataFrame with experiments
        """
        pass

    @abstractmethod
    def validate_candidental(self, values: pd.Series) -> pd.Series:
        """Abstract method to validate the suggested candidates

        Args:
            values (pd.Series): A dataFrame with candidates

        Returns:
            pd.Series: The passed dataFrame with candidates
        """
        pass

    @abstractmethod
    def sample(self, n: int) -> pd.Series:
        """Sample a series of allowed values.

        Args:
            n (int): Number of samples

        Returns:
            pd.Series: Sampled values.
        """
        pass

    @abstractmethod
    def get_bounds(
        self,
        transform_type: Optional[TTransform] = None,
        values: Optional[pd.Series] = None,
    ) -> Tuple[List[float], List[float]]:
        """Returns the bounds of an input feature depending on the requested transform type.

        Args:
            transform_type (Optional[TTransform], optional): The requested transform type. Defaults to None.
            values (Optional[pd.Series], optional): If values are provided the bounds are returned taking
                the most extreme values for the feature into account. Defaults to None.

        Returns:
            Tuple[List[float], List[float]]: List of lower bound values, list of upper bound values.
        """
        pass


class Output(Feature):
    """Base class for all output features.

    Attributes:
        key(str): Key of the Feature.
    """

    @abstractmethod
    def __call__(self, values: pd.Series) -> pd.Series:
        pass

    @abstractmethod
    def validate_experimental(self, values: pd.Series) -> pd.Series:
        """Abstract method to validate the experimental Series

        Args:
            values (pd.Series): A dataFrame with values for the outcome

        Returns:
            pd.Series: The passed dataFrame with experiments
        """
        pass


def is_numeric(s: Union[pd.Series, pd.DataFrame]) -> bool:
    if isinstance(s, pd.Series):
        return pd.to_numeric(s, errors="coerce").notnull().all()
    return s.apply(lambda s: pd.to_numeric(s, errors="coerce").notnull().all()).all()  # type: ignore


def is_categorical(s: pd.Series, categories: List[str]):
    return sum(s.isin(categories)) == len(s)


TInputTransformSpecs = Dict[str, Union[CategoricalEncodingEnum, AnyMolFeatures]]


TDescriptors = Annotated[List[str], Field(min_items=1)]


TCategoryVals = Annotated[List[str], Field(min_items=2)]
TAllowedVals = Optional[Annotated[List[bool], Field(min_items=2)]]


TCategoricalDescriptorVals = Annotated[
    Union[List[List[float]], List[List[int]]],
    Field(min_items=1),
]

TDiscreteVals = Annotated[List[float], Field(min_items=1)]

_CAT_SEP = "_"
